#include <stdio.h>
#include "sctllinux.h"

static char *vkind2str[] = {
	[Vlocal] = "local",
	[Vparam] = "param",
};

static char *reg2str32[] = {
	[ 0] = "eax",
	[ 1] = "ecx",
	[ 2] = "edx",
	[ 3] = "ebx",
	[ 4] = "esp",
	[ 5] = "ebp",
	[ 6] = "esi",
	[ 7] = "edi",
	[ 8] = "eip",
};

static char *reg2str64[] = {
	[ 0] = "rax",
	[ 1] = "rdx",
	[ 2] = "rcx",
	[ 3] = "rbx",
	[ 4] = "rsi",
	[ 5] = "rdi",
	[ 6] = "rbp",
	[ 7] = "rsp",
	[ 8] = "r8",
	[ 9] = "r9",
	[10] = "r10",
	[11] = "r11",
	[12] = "r12",
	[13] = "r13",
	[14] = "r14",
	[15] = "r15",
	[16] = "rip",
	[17] = "xmm0",
	[18] = "xmm1",
	[19] = "xmm2",
	[20] = "xmm3",
	[21] = "xmm4",
	[22] = "xmm5",
	[23] = "xmm6",
	[24] = "xmm7",
	[25] = "xmm8",
	[26] = "xmm9",
	[27] = "xmm10",
	[28] = "xmm11",
	[29] = "xmm12",
	[30] = "xmm13",
	[31] = "xmm14",
	[32] = "xmm15",
	/* st0-st7: floating point registers */
	[33] = "st0",
	[34] = "st1",
	[35] = "st2",
	[36] = "st3",
	[37] = "st4",
	[38] = "st5",
	[39] = "st6",
	[40] = "st7",
	[41] = "mm0",
	[42] = "mm1",
	[43] = "mm2",
	[44] = "mm3",
	[45] = "mm4",
	[46] = "mm5",
	[47] = "mm6",
	[48] = "mm7",
	[49] = "eflags",
	[50] = "es",
	[51] = "cs",
	[52] = "ss",
	[53] = "ds",
	[54] = "fs",
	[55] = "gs",
	[58] = "fs_base",
	[59] = "gs_base",
};

static char*
_fmtlexpr(Lexpr *le, char *reg2str[])
{
	char *p, *o1, *o2;
	u32 m;
	if(le == 0)
		xfatal("unexpected null lexpr");
	switch(le->lkind){
	case Lreg:
		return xstrdup(reg2str[le->no]);
	case Ladd:
	case Lsub:
		o1 = _fmtlexpr(le->op1, reg2str);
		o2 = _fmtlexpr(le->op2, reg2str);
		m = 1+strlen(o1)+1+strlen(o2)+2;
		p = xmalloc(m);
		snprintf(p, m, "(%s%c%s)",
			 o1,
			 le->lkind == Ladd ? '+' : '-',
			 o2);
		free(o1);
		free(o2);
		return p;
	case Lulit:
		m = 20; /* 0x... */
		p = xmalloc(m);
		snprintf(p, m, "0x%lx", le->u);
		return p;
	case Lslit:
		m = 20; /* 0x... */
		p = xmalloc(m);
		snprintf(p, m, "0x%lx", le->s);
		return p;
	default:
		xfatal("unexpected lexpr kind: %d", le->lkind);
		return 0;
	}
}

static char*
fmtlexpr(Lexpr *le, char *reg2str[])
{
	char *p, *o;
	u32 m;
	if(le == 0)
		xfatal("unexpected null lexpr");
	o = _fmtlexpr(le, reg2str);
	if(le->lkind == Lreg)
		return o;
	m = 2+strlen(o)+2;
	p = xmalloc(m);
	snprintf(p, m, "*(%s)", o);
	free(o);
	return p;
}

void
dumpnewlocs(Bin *b, Locs *locs)
{
	u32 i, j;
	Loc *loc;
	L *l;
	char *p;
	for(i = 0; i < locs->n; i++){
		loc = &locs->loc[i];
		p = fmtdecl(loc->type, loc->id);
		printf("\t%s %s\n", vkind2str[loc->vkind], p);
		free(p);
		for(j = 0; j < loc->ls->n; j++){
			l = &loc->ls->l[j];
			if (b->mode == M64)
				p = fmtlexpr(l->le, reg2str64);
			else if (b->mode == M32)
				p = fmtlexpr(l->le, reg2str32);
			else
				return;
			printf("\t\t%lx - %lx %s\n", l->lo, l->hi, p);
			free(p);
		}
	}
}

static void
es(Bin *b, char **err)
{
	Syms *ss;
	Sym *s;
	u32 i;
	Locs *ls;
	Ctx ctx;

	memset (&ctx, 0, sizeof (Ctx));
	ss = enumsym(b);
	if(ss == 0){
		fprintf(stderr, "cannot enum syms\n");
		return;
	}
	for(i = 0; i < ss->n; i++){
		s = ss->sym[i];
		if(s->val > 0 && s->type->kind == Tfun){
			printf("%s\n", s->id);
			if (b->mode == M64)
				ctx.regs.PC = (u64) (s->val);
			else if (b->mode == M32)
				ctx.regs.PC = (u32) (s->val);
			ls = enumloc(b, ctx.regs.PC, err);
			if(ls == 0){
				fprintf(stderr, "cannot find locals for %s\n",
					s->id);
				continue;
			}
			dumpnewlocs(b, ls);
			freelocs(ls);
		}
	}
}

static void
ls(Bin *b, char *id)
{
	Sym *s;
	char *p;
	Type *t, *tn, *ot;
	Decl *d;
	u32 i;
	char *err;

	s = looksym(b, "looksym");
	if(s == 0){
		fprintf(stderr, "cannot find %s\n", id);
		return;
	}
	p = fmttype(s->type);
	printf("type of %s is %s\n", id, p);
	free(p);

	tn = s->type->ds[0]->type->link;
	t = looktype(b, tn, &err);
	if(t == 0){
		p = fmttype(tn);
		fprintf(stderr, "cannot find definition of %s: %s\n", p, err);
		free(p);
		return;
	}
	tn = t->link;
	ot = t;
	t = looktype(b, tn, &err);
	freetype(ot);
	if(t == 0){
		p = fmttype(tn);
		fprintf(stderr, "cannot find definition of %s: %s\n", p, err);
		free(p);
		return;
	}
	p = fmttype(t);
	printf("fields of %s (%d fields):\n", p, t->n);
	free(p);
	for(i = 0; i < t->n; i++){
		d = t->ds[i];
		p = fmtdecl(d->type, d->id);
		printf("\t@0x%x\t%s\n", d->off, p);
		free(p);
	}
	freetype(t);
}

int
main(int argc, char *argv[])
{
	Bin *b;
	char *err;

	if(argc < 2){
		fprintf(stderr, "usage: %s <executable>\n", argv[0]);
		exit(1);
	}
	sctlinit();
	b = names(argv[1], &err);
	if(b == 0){
		fprintf(stderr, "names: %s: %s", argv[1], err);
		exit(1);
	}
	es(b, &err);
	freebin(b);
	sctlfini();
	return 0;
}
