#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../unix/unix.h"
#include "freebsd.h"

#undef XRD
#define XRD(to,fr,sz) { 		\
	if(len < (sz)) goto fail;	\
	memcpy((to),(fr),(sz));		\
	(fr) += (sz);			\
	len -= (sz);			\
}

void
freelaunch(Launch *l)
{
	u64 i;
	for(i = 0; i < l->narg; i++)
		free(l->argv[i]);
	free(l->argv);
	free(l);
}

Launch*
unpacklaunch(void *buf, u64 len)
{
	Launch *l;
	char *p;
	u64 i, m;

	l = xmalloc(sizeof(Launch));
	p = buf;
	XRD(&l->flags, p, sizeof(l->flags));
	XRD(&l->narg, p, sizeof(l->narg));
	l->argv = xmalloc((l->narg+1)*sizeof(char*));
	for(i = 0; i < l->narg; i++){
		XRD(&m, p, sizeof(m));
		l->argv[i] = xmalloc(m+1);
		XRD(l->argv[i], p, m);
	}
	return l;
fail:
	freelaunch(l);
	return 0;
}

Ctl*
launch(Launch *l, char **err)
{
	Ctl *ctl;
	int fd0[2], fd1[2], fd2[2];
	int pid, st;
	char *kerr;

	if(l->flags&Fstdin)
		if(0 > socketpair(PF_UNIX, SOCK_STREAM, 0, fd0))
			xfatalno("socketpair");
	if(l->flags&Fstdout)
		if(0 > socketpair(PF_UNIX, SOCK_STREAM, 0, fd1))
			xfatalno("socketpair");
	if(l->flags&Fstderr)
		if(0 > socketpair(PF_UNIX, SOCK_STREAM, 0, fd2))
			xfatalno("socketpair");

	pid = fork();
	switch(pid){
	case -1:
		if(err)
			*err = strerror(errno);
		goto fail;
	case 0:
		if(l->flags&Fstdin){
			close(fd0[0]);
			dup2(fd0[1], 0);
			if(fd0[1] != 0)
				close(fd0[1]);
		}
		if(l->flags&Fstdout){
			close(fd1[0]);
			dup2(fd1[1], 1);
			if(fd1[1] != 1)
				close(fd1[1]);
		}
		if(l->flags&Fstderr){
			close(fd2[0]);
			dup2(fd2[1], 2);
			if(fd2[1] != 2)
				close(fd2[1]);
		}
		if(0 > ptrace(PT_TRACE_ME, 0, 0, 0))
			xfatalno("PT_TRACE_ME");
		execv(l->argv[0], l->argv);
		exit(1);
	}

	if(0 > waitpid(pid, &st, 0))
		xfatalno("waitpid");

	if(WIFEXITED(st)
	   || WIFSIGNALED(st)
	   || (WIFSTOPPED(st) && WSTOPSIG(st) != SIGTRAP)){
		if(err)
			*err = "cannot exec";
		goto fail;
	}

	ctl = mkctl(pid);
	ctl->state = Sstop;
	ctl->whence = Wlaunched;
	ctl->as = mkas(ctl);
	xgetregs(ctl, &ctl->ctx.regs);
	if(0 > syminitexec(ctl, err)){
		ctlkill(ctl, &kerr);
		goto fail;
	}
	kqtrace(pid);
	if(l->flags&Fstdin){
		ctl->fd[0] = fd0[0];
		close(fd0[1]);
	}
	if(l->flags&Fstdout){
		ctl->fd[1] = fd1[0];
		close(fd1[1]);
	}
	if(l->flags&Fstderr){
		ctl->fd[2] = fd2[0];
		close(fd2[1]);
	}

	return ctl;
fail:
	if(l->flags&Fstdin){
		close(fd0[0]);
		close(fd0[1]);
	}
	if(l->flags&Fstdout){
		close(fd1[0]);
		close(fd1[1]);
	}
	if(l->flags&Fstderr){
		close(fd2[0]);
		close(fd2[1]);
	}
	return 0;
}
