#include "../inc/sctl.h"

static int kq;

void
kqtrace(int pid)
{
	struct kevent kev;
	memset(&kev, 0, sizeof(kev));
	EV_SET(&kev, pid, EVFILT_PROC, EV_ADD,
	       NOTE_FORK|NOTE_TRACK, 0, 0);
	if(0 > kevent(kq, &kev, 1, 0, 0, 0))
		xfatalno("kevent EVFILT_PROC ADD");
}

void
kqnotrace(int pid)
{
	struct kevent kev;
	EV_SET(&kev, pid, EVFILT_PROC, EV_DELETE, 0, 0, 0);
	if(0 > kevent(kq, &kev, 1, 0, 0, 0)){
		if(errno != ENOENT)
			xfatalno("kevent EVFILT_PROC DELETE");
	}
}

void
srv(int fd)
{
	int rv;
	struct kevent kev[2];
	Msg *m;
	Event *e;

	xdebug(Daction, "%s()", __FUNCTION__);
	sctlinit();
	cloexec(fd);
	signal(SIGCHLD, SIG_IGN);
	kq = kqueue();
	if(0 > kq)
		xfatalno("kqueue");
	EV_SET(&kev[0], SIGCHLD, EVFILT_SIGNAL, EV_ADD, 0, 0, 0);
	EV_SET(&kev[1], fd, EVFILT_READ, EV_ADD, 0, 0, 0);
	if(0 > kevent(kq, kev, 2, 0, 0, 0))
		xfatalno("kevent");
	while(1){
		rv = kevent(kq, 0, 0, kev, 1, 0);
		if(0 > rv)
			xfatalno("kevent");
		if(rv != 1)
			xfatal("kevent: expected 1, got %d", rv);
		switch(kev[0].filter){
		case EVFILT_READ:
			m = readmsg(fd);
			if(m == 0){
				closechan(fd);
				ctlrelease();
				sctlfini();
				killall();
			}
			domsg(fd, m);
			freemsg(m);
			break;
		case EVFILT_SIGNAL:
			/* SIGCHLD interruption implies
			   that one or more events are
			   waiting */
			while((e = ctlpoll(0))){
				aevent(fd, e);
				freeevent(e);
			}
			break;
		case EVFILT_PROC:
			xlog("EVFILT_PROC %d %x %x %x",
			     kev[0].ident, kev[0].flags, kev[0].fflags,
			     kev[0].data);
			if(kev[0].fflags&NOTE_FORK)
				xlog("%d called fork", kev[0].ident);
			if(kev[0].fflags&NOTE_TRACK)
				xlog("%d forked", kev[0].ident);
			if(kev[0].fflags&NOTE_CHILD)
				xlog("%d is a child of %d",
				     kev[0].ident, kev[0].data);
			break;
		default:
			xfatal("kevent: unexpected filter: %d", kev[0].filter);
		}
	}
}
