#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/socket.h>
#include <sys/event.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <machine/reg.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <signal.h>
#include <unistd.h>

typedef int8_t		s8;
typedef int16_t		s16;
typedef int32_t		s32;
typedef int64_t		s64;
typedef uint8_t		u8;
typedef uint16_t	u16;
typedef uint32_t	u32;
typedef uint64_t	u64;
typedef uintptr_t	uptr;

typedef struct reg Reg;
typedef int Fd;
typedef int Chan;
typedef struct Bin Bin;

typedef
struct Xas
{
	void *aux;		/* auxiliary vector */
	void *scratch;		/* breakpoint step scratch space */
	Bin *bin;		/* binaries loaded into address space */
	Bin *ld;		/* loader binary, if present */
	unsigned char mode;	/* M32 or M64 */
} Xas;

typedef
struct Xctl
{
} Xctl;

typedef
struct Launch
{
	u64 flags;
	u64 narg;
	char **argv;
} Launch;

typedef
struct Thread
{
	int tid;
} Thread;

typedef
struct Proc
{
	int pid;
	char *cmd;
	u64 nth;
	Thread *th;
} Proc;

typedef
struct Ps
{
	u64 npr;
	Proc *pr;
} Ps;

typedef
struct Region
{
	char *file;
	u64 addr;
	u64 len;
	u32 flags;
} Region;

typedef
struct Stat
{
	int pid;
	u64 nr;
	Region *r;
	u64 nb;
	Bin **b;
} Stat;

typedef
struct Ctx
{
	Reg regs;
	int cursig;
} Ctx;

typedef
enum Evkind
{
	Eclear	 = 0,
	Eset	 = 1,
	Esyscall = (Eset<<1),
	Eexec	 = (Esyscall<<1),
	Efork    = (Eexec<<1),
	Eclone   = (Efork<<1),
	Esignal  = (Eclone<<1),
	Eexit    = (Esignal<<1),
	Etrap    = (Eexit<<1),
	Esnap    = (Etrap<<1),
	Estepctx = (Esnap<<1),
	Eload    = (Estepctx<<1),
	Eunload  = (Eload<<1),
} Evkind;

typedef struct Event Event;
struct Event
{
	Event *link;
	Evkind kind;
	u64 id;
	union {
		struct syscall {
			Ctx ctx;
		} syscall;
		struct trap {
			u64 id;
			Ctx ctx;
		} trap;
		struct snap {
			u64 id;
			u64 cid;
			Ctx cctx;
		} snap;
		struct spawn {
			u64 cid;
			Ctx ctx;
			Ctx cctx;
		} spawn;
		struct exec {
			Ctx ctx;
		} exec;
		struct signal {
			Ctx ctx;
			siginfo_t info;
		} signal;
		struct exit {
		} exit;
		struct load {
			Bin **bs;
			Ctx ctx;
		} load;
		struct unload {
		} unload;
	} u; /* for c++ */
};

#if defined(__amd64__)
#define PC r_rip
#define SP r_rsp
#define DI r_rdi
#define SI r_rsi
#else
#define PC r_eip
#define SP r_esp
#define DI r_edi
#define SI r_esi
#endif
