#include "../inc/sctl.h"
#include "../inc/err.h"

typedef
struct Packed
{
	u64 r15;
	u64 r14;
	u64 r13;
	u64 r12;
	u64 rbp;
	u64 rbx;
	u64 r11;
	u64 r10;
	u64 r9;
	u64 r8;
	u64 rax;
	u64 rcx;
	u64 rdx;
	u64 rsi;
	u64 rdi;
	u64 orig_rax;
	u64 rip;
	u64 cs;
	u64 eflags;
	u64 rsp;
	u64 ss;
	u64 fs_base;
	u64 gs_base;
	u64 ds;
	u64 es;
	u64 fs;
	u64 gs;
	u64 cursig;
} Packed;

u64
ctlpc(Ctl *ctl)
{
	return ctl->ctx.regs.r_eip;
}

void
ctlsetpc(Ctl *ctl, u64 pc)
{
	ctl->ctx.regs.r_eip = pc;
}

void
freectx(Ctx *c)
{
	free(c);
}

Ctx*
unpackctx(void *bytes, u64 n)
{
	Packed *p;
	Ctx *ctx;

	if(n != sizeof(Packed))
		return 0;
	p = bytes;
	ctx = xmalloc(sizeof(Ctx));
	ctx->regs.r_ebx = p->rbx;
	ctx->regs.r_ecx = p->rcx;
	ctx->regs.r_edx = p->rdx;
	ctx->regs.r_esi = p->rsi;
	ctx->regs.r_edi = p->rdi;
	ctx->regs.r_ebp = p->rbp;
	ctx->regs.r_eax = p->rax;
	ctx->regs.r_ds = p->ds;
	ctx->regs.r_es = p->es;
	ctx->regs.r_fs = p->fs;
	ctx->regs.r_gs = p->gs;
	ctx->regs.r_eip = p->rip;
	ctx->regs.r_cs = p->cs;
	ctx->regs.r_eflags = p->eflags;
	ctx->regs.r_esp = p->rsp;
	ctx->regs.r_ss = p->ss;

	// FIXME: what are these?
	ctx->regs.r_isp = 0;
	ctx->regs.r_trapno = 0;
	ctx->regs.r_err = 0;

	return ctx;
}

void
packctx(Buf *ob, Ctx *ctx)
{
	Packed p;

	memset(&p, 0, sizeof(Packed));

	p.rbx = ctx->regs.r_ebx;
	p.rcx = ctx->regs.r_ecx;
	p.rdx = ctx->regs.r_edx;
	p.rsi = ctx->regs.r_esi;
	p.rdi = ctx->regs.r_edi;
	p.rbp = ctx->regs.r_ebp;
	p.rax = ctx->regs.r_eax;
	p.ds = ctx->regs.r_ds;
	p.es = ctx->regs.r_es;
	p.fs = ctx->regs.r_fs;
	p.gs = ctx->regs.r_gs;
	p.rip = ctx->regs.r_eip;
	p.cs = ctx->regs.r_cs;
	p.eflags = ctx->regs.r_eflags;
	p.rsp = ctx->regs.r_esp;
	p.ss = ctx->regs.r_ss;

	bufpush(ob, &p, sizeof(Packed));
}
