#include "../inc/sctl.h"
#include "../inc/err.h"

typedef
struct Packed
{
	u64 r15;
	u64 r14;
	u64 r13;
	u64 r12;
	u64 rbp;
	u64 rbx;
	u64 r11;
	u64 r10;
	u64 r9;
	u64 r8;
	u64 rax;
	u64 rcx;
	u64 rdx;
	u64 rsi;
	u64 rdi;
	u64 orig_rax;
	u64 rip;
	u64 cs;
	u64 eflags;
	u64 rsp;
	u64 ss;
	u64 fs_base;
	u64 gs_base;
	u64 ds;
	u64 es;
	u64 fs;
	u64 gs;
	u64 cursig;
} Packed;

u64
ctlpc(Ctl *ctl)
{
	return ctl->ctx.regs.r_rip;
}

void
ctlsetpc(Ctl *ctl, u64 pc)
{
	ctl->ctx.regs.r_rip = pc;
}

void
freectx(Ctx *c)
{
	free(c);
}

Ctx*
unpackctx(void *bytes, u64 n)
{
	Packed *p;
	Ctx *ctx;

	if(n != sizeof(Packed))
		return 0;
	p = bytes;
	ctx = xmalloc(sizeof(Ctx));
	ctx->regs.r_rbx = p->rbx;
	ctx->regs.r_rcx = p->rcx;
	ctx->regs.r_rdx = p->rdx;
	ctx->regs.r_rsi = p->rsi;
	ctx->regs.r_rdi = p->rdi;
	ctx->regs.r_rbp = p->rbp;
	ctx->regs.r_rax = p->rax;
	ctx->regs.r_rip = p->rip;
	ctx->regs.r_cs = p->cs;
	ctx->regs.r_rflags = p->eflags;
	ctx->regs.r_rsp = p->rsp;
	ctx->regs.r_ss = p->ss;

	// FIXME: what are these?
	ctx->regs.r_trapno = 0;
	ctx->regs.r_err = 0;

	return ctx;
}

void
packctx(Buf *ob, Ctx *ctx)
{
	Packed p;

	memset(&p, 0, sizeof(Packed));

	p.rbx = ctx->regs.r_rbx;
	p.rcx = ctx->regs.r_rcx;
	p.rdx = ctx->regs.r_rdx;
	p.rsi = ctx->regs.r_rsi;
	p.rdi = ctx->regs.r_rdi;
	p.rbp = ctx->regs.r_rbp;
	p.rax = ctx->regs.r_rax;
	p.rip = ctx->regs.r_rip;
	p.cs = ctx->regs.r_cs;
	p.eflags = ctx->regs.r_rflags;
	p.rsp = ctx->regs.r_rsp;
	p.ss = ctx->regs.r_ss;

	bufpush(ob, &p, sizeof(Packed));
}
