enum
{
	Mbrk	= 0,
	Msnap	= 1,
	Mld	= 2,
	Mentry	= 3,
};

enum
{
	// execution modes
	M32,	// 32-bit
	M64,	// 64-bit
	Mle,	/* little-endian */
	Mbe,	/* big-endian */
};

typedef
enum State
{
	Srun	= 1,
	Sstop,
	Strap,
	Sdead,
} State;

typedef
enum Whence
{
	Wunknown,
	Wattached,
	Wlaunched,
	Wsnapped,
} Whence;

typedef struct As As;
typedef struct Brk Brk;
typedef struct Brkref Brkref;
typedef struct Ctl Ctl;

/* breakpoint */
struct Brk
{
	u64 mode;
	u64 addr;
	u32 cnt;		/* reference count */
	u8 text;		/* original text */
};

/* breakpoint reference */
struct Brkref
{
	u32 id;
	Brk *bp;
	Brkref *link;		/* per-ctl brkref list for this bp */
};

/* address space */
struct As
{
	Ctl *ctl;		/* ctls in this address space */
	HT *bpaddr;		/* breakpoints by addr */
	Xas x;			/* system-specific state */
};

/* thread of control */
struct Ctl
{
	u64 id;
	As *as;
	Fd fd[3];		/* redirected fds */
	State state;		/* Running, Stopped, ... */
	Whence whence;		/* Attached, Launched */
	Ctx ctx;		/* registers and other context */
	Evkind tracemode;
	Brk curbrk;		/* Atbreak breakpoint */
	Ctl *link;		/* address space list */
	u32 nextid;		/* next breakpoint ref identifier */
	HT *braddr;		/* breakpoint refs by addr */
	HT *brid;		/* breakpoint refs by id */
	Xctl x;			/* system-specific state */
};

/* trap request */
typedef
struct Trap
{
	u64 mode;
	u64 addr;
} Trap;

typedef
struct Trace
{
	u64 flags;
} Trace;

/* port/prctl.c */
As*		_mkas(Ctl *ctl);
Ctl*		_mkctl(u64 id);
void		_freeas(As *as);
void		_freectl(Ctl *ctl);
void		_dead(Ctl *ctl);
int		_ctltrace(Ctl *ctl, Trace *t, char **err);
void		ctladd(As *as, Ctl *ctl);
void		ctldel(As *as, Ctl *ctl);
void		ctlrelease();
char*		ctlreads(Ctl *ctl, u64 a, char **err);
Brkref*		atbrk(Ctl *ctl);
int		clrbrkref(Ctl *ctl, Brkref *br, char **err);
void		linkspawn(Ctl *ctl, Ctl *spwn, int copy);
Ctl*		lookctl(u64 id);
void		ABUSE(u64 id, Ctl *ctl);
void		_sctlfini();
void		_sctlinit();
int		steptrap(Ctl *ctl, char **err);

/* platform-dependent */
extern char	osversion[];
As*		mkas(Ctl *ctl);
void		freeas(As *as);
void		freectl(Ctl *ctl);
void		dead(Ctl *ctl);
int		setbrk(Ctl *ctl, Brk *bp, char **err);
int		clrbrk(Ctl *ctl, Brk *bp, char **err);
u64		ctlpc(Ctl *ctl);
void		ctlsetpc(Ctl *ctl, u64 pc);
void		symdupas(As *as, As *nas);

/* API */
Ctl*		launch(Launch *l, char **err);
Ctl*		attach(int pid, char **err);
Ps*		ps(char **err);
Brkref*		ctlsettrap(Ctl *ctl, Trap *t, char **err);
int		ctlclrtrap(Ctl *ctl, u32 id, char **err);
int		ctltrace(Ctl *ctl, Trace *t, char **err);
Ctx*		ctlgetctx(Ctl *ctl, char **err);
int		ctlsetctx(Ctl *ctl, Ctx *c, char **err);
Ctl*		ctlsnap(Ctl *ctl);
Stat*		ctlstat(Ctl *ctl, char **err);
int		ctlread(Ctl *ctl, unsigned long addr, void *buf,
			unsigned long cnt, char **err);
int		ctlwrite(Ctl *ctl, unsigned long addr, void *buf,
			 unsigned long cnt, char **err);
int		ctlstep(Ctl *ctl, char **err);
void		ctlcont(Ctl *ctl);
Event*		ctlstop(Ctl *ctl, char **err);
int		ctldetach(Ctl *ctl, char **err);
int		ctlkill(Ctl *ctl, char **err);
Event*		ctlpoll(Ctl *maybectl);
Event*		ctlwait(Ctl *maybectl);
void		sctlfini();
void		sctlinit();


/* not ready for prime time */
int		ctlns(u8 op, void *buf, u64 len, void *xbuf, char **err);
