typedef
enum Tkind
{
	Tvoid = 0,
	Tbase = 1,
	Tstruct = 2,
	Tunion = 3,
	Tenum = 4,
	Tptr = 5,
	Tarr = 6,
	Tfun = 7,
	Ttypedef = 8,
	Tbitfield = 9,
	Tconst = 10,
	Txaccess = 11,		/* unused */
	Tundef = 12,
} Tkind;

// FIXME: distinguish signed char from char
typedef
enum Cbase
{
	Vundef=0,
	Vlo,
	Vbool=Vlo,
	Vchar,
	Vshort,
	Vint,
	Vlong,
	Vvlong,
	Vuchar,
	Vushort,
	Vuint,
	Vulong,
	Vuvlong,
	Vfloat,
	Vdouble,
	Vlongdouble,
	Vcomplex,
	Vdoublex,
	Vlongdoublex,
	Vnbase,
	Vptr = Vnbase,		/* alias for some other base type */
	Vvoid,
	Vnallbase,
} Cbase;

/* base type representations */
typedef
enum Rkind
{
	Rundef,
	Ru08le,
	Ru16le,
	Ru32le,
	Ru64le,
	Rs08le,
	Rs16le,
	Rs32le,
	Rs64le,
	Ru08be,
	Ru16be,
	Ru32be,
	Ru64be,
	Rs08be,
	Rs16be,
	Rs32be,
	Rs64be,
	Rf32,
	Rf64,
	Rf96,
	Rf128,
	Rx64,
	Rx128,
	Rx192,
	Rnrep,
} Rkind;

typedef struct Enum Enum;
typedef struct Type Type;
typedef struct Decl Decl;
typedef struct Syms Syms;

struct Decl
{
	Type *type;
	char *id;		/* field/param name */
	u32 off;		/* field offset */
};

struct Enum
{
	char *id;
	u64 val;
	Type *type;		/* defining enum */
};

struct Type
{
	Tkind kind;
	Cbase base;		/* base type */
	Rkind rep;		/* base */
	char *id;		/* typedef, struct, union, enum */
	u32 sz;			/* struct/union size */
	u32 n;			/* array # elements;
				   struct/union # fields;
				   function # params;
				   enum # consts */
	u32 pos;		/* bitfield offset */
	u32 w;			/* bitfield width */
	Decl **ds;		/* struct/union fields; func params */
	Enum **en;		/* enum constants */
	Type *link;		/* typedef, ptr, arr, func, const */
};

typedef
struct Types
{
	Type **t;
	u32 n, max;
} Types;

enum Stype
{
	Sundef = 1,
	Sdata = Sundef<<1,
	Stext = Sdata<<1,
	Sro = Stext<<1,
	Senum = Sro<<1,
	Sinline = Senum<<1,
};

typedef
struct Sym
{
	char *id;
	u8 flags;
	u64 val;
	u64 sz;
	Type *type;

} Sym;

struct Syms
{
	Sym **sym;
	u32 n, max;
};

typedef
enum Lkind {
	Lreg,
	Lderef,
	Ladd,
	Lsub,
	Lulit,
	Lslit,
} Lkind;

typedef struct Lexpr Lexpr;
struct Lexpr
{
	Lkind lkind;
	u32 no;			/* Lreg */
	u64 u;			/* Lulit */
	s64 s;			/* Lslit */
	Lexpr *op1, *op2;
};

typedef
struct L
{
	u64 lo, hi;
	Lexpr *le;
} L;

typedef
struct Ls
{
	L *l;
	u32 n, max;
} Ls;

typedef
enum Vkind
{
	Vlocal,
	Vparam,
} Vkind;

typedef
struct Loc
{
	char *id;
	u64 sz;
	Type *type;
	Vkind vkind;
	Ls *ls;
} Loc;

typedef
struct Locs
{
	u64 nsid;
	Loc *loc;
	u32 n, max;
} Locs;

void		addtype(Types *ts, Type *t);
Decl*		copydecl(Decl *d);
Enum*		copyenum(Enum *e, Type *t);
Type*		copytype(Type *t);
char*		fmtdecl(Type *t, char *id);
char*		fmttype(Type *t);
void		freedecl(Decl *d);
void		freeenumconst(Enum *e);
void		freetype(Type *t);
void		freetypes(Types *ts);
Decl*		mkdecl();
Type*		mkenum(char *tag);
Enum*		mkenumconst();
Type*		mkptr(Type *st);
Type*		mkstruct(char *tag);
Type*		mksue(Tkind kind, char *tag);
Type*		mktypedef(char *tid);
Type*		mkunion(char *tag);
Type*		mktype();
Types*		mktypes();
Type*		mkvoid();
