#include "../inc/sctl.h"
#include "../inc/err.h"

enum
{
	Maxpath = 128,
};

static int
isint(char *s)
{
	while(*s){
		if(!isdigit(*s))
			return 0;
		s++;
	}
	return 1;
}

static int
parsestat(Proc *pr, char **err)
{
	char path[Maxpath];
	char *cmd, *buf, *p;

	snprintf(path, sizeof(path), "/proc/%d/stat", pr->pid);
	buf = readfile(path, 0);
	if(buf == 0){
		if(err)
			*err = lasterror();
		return -1;
	}

	/* pid */
	strtoul(buf, &p, 10);
	if(p == buf)
		goto strange;

	/* cmd */
	p = strchr(p, '(');
	if(p == 0)
		goto strange;
	p++;
	cmd = p;
	p = strchr(p, ')');
	if(p == 0)
		goto strange;
	*p++ = 0;
	pr->cmd = xstrdup(cmd);
	p++;
	free(buf);
	return 0;
strange:
	free(buf);
	if(err)
		*err = Estrange;
	return -1;
}

static int
threads(Proc *pr, char **err)
{
	DIR *d;
	char path[Maxpath];
	struct dirent *e;
	Thread *th;
	u64 max;

	max = 128;
	pr->th = xmalloc(max*sizeof(Thread));

	snprintf(path, sizeof(path), "/proc/%d/task", pr->pid);
	d = opendir(path);
	if(d == 0){
		if(err)
			*err = Estrange;
		return -1;
	}
	while(1){
		errno = 0;
		e = readdir(d);
		if(e == 0 && errno != 0){
			if(err)
				*err = lasterror();
			goto fail;
		}
		if(e == 0)
			break;
		if(!isint(e->d_name))
			continue;

		if(pr->nth >= max){
			xfatal("this seems to be a buggy allocation");
			pr->th = xrealloc(pr->th, max, 2*max);
			max *= 2;
		}
		th = &pr->th[pr->nth++];
		th->tid = strtoul(e->d_name, 0, 10);
	}
	closedir(d);
	return 0;
fail:
	closedir(d);
	return -1;
}

void
packps(Buf *b, Ps *p)
{
	u64 i, j;
	Proc *pr;
	Thread *th;

	bufpushu64(b, p->npr);
	for(i = 0; i < p->npr; i++){
		pr = &p->pr[i];
		bufpushu64(b, pr->pid);
		bufpushs(b, pr->cmd);
		bufpushu64(b, pr->nth);
		for(j = 0; j < pr->nth; j++){
			th = &pr->th[j];
			bufpushu64(b, th->tid);
		}
	}
}

void
freeps(Ps *p)
{
	u64 i;
	for(i = 0; i < p->npr; i++){
		free(p->pr[i].cmd);
		free(p->pr[i].th);
	}
	free(p->pr);
	free(p);
}

Ps*
ps(char **err)
{
	Ps *p;
	Proc *pr;
	DIR *d;
	struct dirent *e;
	char path[Maxpath];
	int fd;
	u64 n, max;

	d = opendir("/proc");
	if(d == 0)
		return 0;
	n = 0;
	p = xmalloc(sizeof(Ps));
	max = 128;
	p->pr = xmalloc(max*sizeof(Proc));

	while(1){
		errno = 0;
		e = readdir(d);
		if(e == 0 && errno != 0){
			if(err)
				*err = lasterror();
			goto fail;
		}
		if(e == 0)
			break;

		/* skip non-process entries */
		if(!isint(e->d_name))
			continue;

		/* can we access the process? */
		snprintf(path, sizeof(path), "/proc/%s/mem", e->d_name);
		fd = open(path, O_RDONLY);
		if(0 > fd)
			continue;
		close(fd);

		if(p->npr >= max){
			p->pr = xrealloc(p->pr, max*sizeof(Proc),
					 2*max*sizeof(Proc));
			max *= 2;
		}
		pr = &p->pr[p->npr++];
		pr->pid = strtoul(e->d_name, 0, 10);
		if(0 > parsestat(pr, err))
			goto fail;
		if(0 > threads(pr, err))
			goto fail;
	}

	closedir(d);
	return p;
fail:
	freeps(p);
	closedir(d);
	return 0;
}
