#include "../inc/sctl.h"
#include "../unix/unix.h"
#include "linux.h"

static int
snap(Ctl *ctl)
{
	extern char foo[], endfoo[];
	Reg regs, oregs, xregs;
	unsigned long codeaddr, codelen;
	int status, sig, w;
	char *origcode;
	pid_t cpid, pid;

	pid = ctl->id;
	codeaddr = (unsigned long)ctlscratch(ctl, 0);
	if(codeaddr == 0)
		xfatalno("no scratch");

	if(0 > ptrace(PTRACE_GETREGS, pid, 0, &oregs))
		xfatalno("PTRACE_GETREGS");
	regs = oregs;

	codelen = endfoo-foo;
	origcode = xmalloc(codelen);
	if(0 > ctlread(ctl, codeaddr, origcode, codelen, 0))
		xfatalno("ctlread");
	if(0 > ctlwrite(ctl, codeaddr, foo, codelen, 0))
		xfatalno("ctlwrite");

	regs.PC = codeaddr;

	if(0 > ptrace(PTRACE_SETREGS, pid, 0, &regs))
		xfatalno("PTRACE_SETREGS");
	/*
	  there is, of course, no way to tell what options are set
	  already, so we assume that PTRACE_O_TRACEFORK is set
	  among them, rather than clobber the others.
	if(0 > ptrace(PTRACE_SETOPTIONS, pid, 0, PTRACE_O_TRACEFORK))
		xfatalno("PTRACE_SETOPTIONS");
	 */
	if(0 > ptrace(PTRACE_CONT, pid, 0, 0))
		xfatalno("PTRACE_CONT");

	cpid = 0;
	while(1){
		if(0 > waitpid(pid, &status, 0))
			xfatalno("waitpid");
		if(!WIFSTOPPED(status))
			xfatal("wait: not stopped");
		sig = WSTOPSIG(status);
		if(sig == SIGCHLD){
			/* xlog("supressing SIGCHLD"); */
			if(0 > ptrace(PTRACE_CONT, pid, 0, 0))
				xfatalno("PTRACE_CONT");
			continue;
		}
		if(sig != SIGTRAP){
			/*
			 * task got a signal we didn't intend;
			 * allow it and hope for the best.
			 */
			if(0){
				if(0 > ptrace(PTRACE_GETREGS, pid, 0, &xregs))
					xfatalno("PTRACE_GETREGS");
				xlog("fowarding signal %d to parent at %x",
				     sig, xregs.PC);
			}
			if(0 > ptrace(PTRACE_CONT, pid, 0, sig))
				xfatalno("PTRACE_CONT");
			continue;
		}
		/* first SIGTRAP: fork notification */
		if(WSTOPEVENT(status) == PTRACE_EVENT_FORK){
			if(cpid)
				xfatal("unexpected PTRACE_EVENT_FORK");
			if(0 > ptrace(PTRACE_GETEVENTMSG, pid, 0, &cpid))
				xfatalno("PTRACE_GETEVENTMSG");
			if(0)
				xlog("PTRACE_EVENT_FORK: %d", cpid);
			w = waitpid(cpid, 0, 0);
			if(cpid != w)
				xfatalno("expected waitpid behavior (%d)", w);
			if(0 > ptrace(PTRACE_CONT, pid, 0, 0))
				xfatalno("PTRACE_CONT");
			continue;
		}

		/* second SIGTRAP: fork returned in parent */
		if(0)
			xlog("fork returned in parent");
		if(cpid == 0)
			xfatal("fork never happened");
		break;
	}

	/* clean up parent */
	if(0 > ctlwrite(ctl, codeaddr, origcode, codelen, 0))
		xfatalno("ctlwrite");
	if(0 > ptrace(PTRACE_SETREGS, pid, 0, &oregs))
		xfatalno("PTRACE_SETREGS");

	/* wait for child to close fds */
	if(0 > ptrace(PTRACE_CONT, cpid, 0, 0))
		xfatalno("PTRACE_CONT");
	while(1){
		if(0 > waitpid(cpid, &status, 0))
			xfatalno("waitpid");
		if(!WIFSTOPPED(status))
			xfatal("wait: not stopped");
		sig = WSTOPSIG(status);
		if(sig != SIGTRAP){
			/*
			 * task got a signal we didn't intend;
			 * allow it and hope for the best.
			 */
			if(0){
				if(0 > ptrace(PTRACE_GETREGS, cpid, 0, &xregs))
					xfatalno("PTRACE_GETREGS");
				xlog("fowarding signal %d to child at %x",
				     sig, xregs.PC);
			}
			if(0 > ptrace(PTRACE_CONT, cpid, 0, sig))
				xfatalno("PTRACE_CONT");
			continue;
		}
		break;
	}

	/* clean up child */
	if(0 > ctlwrite(ctl, codeaddr, origcode, codelen, 0))
		xfatalno("ctlwrite");
	if(0 > ptrace(PTRACE_SETREGS, cpid, 0, &oregs))
		xfatalno("PTRACE_SETREGS");
	free(origcode);
	return cpid;
}

Ctl*
ctlsnap(Ctl *ctl)
{
	int cid, wasrunning;
	Ctl *spwn;
	char *err;

	wasrunning = 0;
	if(ctl->state == Srun){
		wasrunning = 1;
		ctlstop(ctl, &err);
		if(err)
			return 0;
	}
	cid = snap(ctl);
	if(wasrunning)
		ctlcont(ctl);
	spwn = mkctl(cid);
	spwn->state = Sstop;
	spwn->whence = Wsnapped;
	xgetregs(spwn, &spwn->ctx.regs);
	linkspawn(ctl, spwn, 1);
	return spwn;
}
