#include "../inc/sctl.h"
#include "../inc/err.h"
#include "linux.h"

enum
{
	Maxpath = 128,
};

static int
parseregion(char *line, Region *r, char **err)
{
	char *p;
	unsigned long endaddr;

	r->addr = strtoul(line, &p, 16);
	p++;  /* skip '-' */
	endaddr = strtoul(p, &p, 16);
	r->len = endaddr - r->addr;
	p++;  /* skip ' ' */
	r->flags = 0;
	while(*p != ' ')
		switch(*p++){
		case 'r':
			r->flags |= PROT_READ;
			break;
		case 'w':
			r->flags |= PROT_WRITE;
			break;
		case 'x':
			r->flags |= PROT_EXEC;
			break;
		case '-':
		case 'p':
		case 's': /* "may share" */
			/* Do nothing */
			break;
		default:
			if(err)
				*err = Estrange;
			return -1;
		}

	p = strrchr(p, ' ');
	if(p == 0){
		/* FIXME: does this happen? */
		r->file = xstrdup("");
		return 0;
	}
	p++;
	r->file = xstrdup(p);
	return 0;
}

int
parsemaps(Stat *s, char **err)
{
	u64 n;
	char path[Maxpath];
	char *p, *m, *e;

	snprintf(path, sizeof(path), "/proc/%d/maps", s->pid);
	m = readfile(path, 0);
	if(m == 0){
		if(err)
			*err = lasterror();
		return -1;
	}
	n = 0;
	p = m;
	while(*p){
		if(*p == '\n'){
			n++;
			*p = 0;
		}
		p++;
	}
	e = p;

	s->nr = 0;
	s->r = xmalloc(n*sizeof(Region));
	p = m;
	while(p < e){
		if(0 > parseregion(p, &s->r[s->nr++], err))
			goto fail;
		p += strlen(p)+1;
	}
	free(m);
	return 0;
fail:
	free(m);
	return -1;
}
