#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../unix/unix.h"
#include "linux.h"

enum
{
	Maxpath = 128,
};

struct link_map32
{
	u32	l_addr;
	u32	l_name;
	u32	l_ld;
	u32	l_next, l_prev;
};

struct link_map64
{
	u64	l_addr;
	u64	l_name;
	u64	l_ld;
	u64	l_next, l_prev;
};

/* /usr/include/link.h */
enum {
	/* This state value describes the mapping change taking place when
	   the `r_brk' address is called.  */
	RT_CONSISTENT,		/* Mapping change is complete.  */
	RT_ADD,			/* Beginning to add a new object.  */
	RT_DELETE		/* Beginning to remove an object mapping.  */
} r_state;

char* auxname[] = {
	[AT_NULL] = "AT_NULL",
	[AT_IGNORE] = "AT_IGNORE",
	[AT_EXECFD] = "AT_EXECFD",
	[AT_PHDR] = "AT_PHDR",
	[AT_PHENT] = "AT_PHENT",
	[AT_PHNUM] = "AT_PHNUM",
	[AT_PAGESZ] = "AT_PAGESZ",
	[AT_BASE] = "AT_BASE",
	[AT_FLAGS] = "AT_FLAGS",
	[AT_ENTRY] = "AT_ENTRY",
	[AT_NOTELF] = "AT_NOTELF",
	[AT_UID] = "AT_UID",
	[AT_EUID] = "AT_EUID",
	[AT_GID] = "AT_GID",
	[AT_EGID] = "AT_EGID",
	[AT_PLATFORM] = "AT_PLATFORM",
	[AT_HWCAP] = "AT_HWCAP",
	[AT_CLKTCK] = "AT_CLKTCK",
	[AT_SECURE] = "AT_SECURE",
	[AT_BASE_PLATFORM] = "AT_BASE_PLATFORM",
	[AT_RANDOM] = "AT_RANDOM",
	[AT_EXECFN] = "AT_EXECFN",
	[AT_SYSINFO_EHDR] = "AT_SYSINFO_EHDR",
};

/* the bs array is zero-terminated */
static void
addbs(Bin ***bsp, unsigned *nbsp, unsigned *mbsp, Bin *b)
{
	Bin **bs;
	unsigned nbs, mbs;

	bs = *bsp;
	nbs = *nbsp;
	mbs = *mbsp;

	if(mbs == 0){
		mbs = 4;
		bs = xmalloc(mbs*sizeof(Bin*));
		nbs = 0;
		bs[nbs++] = b;
	}else if(nbs >= mbs-1){
		bs = xrealloc(bs, mbs*sizeof(Bin*), 2*mbs*sizeof(Bin*));
		mbs *= 2;
		bs[nbs++] = b;
	}else
		bs[nbs++] = b;

	*bsp = bs;
	*nbsp = nbs;
	*mbsp = mbs;
}

/* this function is a copy of scanrmap64 with 32-bit types
   for LMP and LM. */
static int
scanrmap32(Ctl *ctl, Bin ***bs, char **err)
{
	u32 lmp;
	char *path;
	struct link_map32 lm;
	Bin *b;
	unsigned nbs, mbs;
	
	nbs = mbs = 0;
	*bs = 0;
	if(0 > ctlread(ctl, ctl->as->x.r_map, &lmp, sizeof(lmp), err))
		return -1;
	while(lmp){
		if(0 > ctlread(ctl, lmp, &lm, sizeof(lm), err))
			return -1;
		path = ctlreads(ctl, lm.l_name, err);
		if(path == 0)
			return -1;
		if(strlen(path) == 0)
			goto again;
		if(!lookbinpath(ctl, path)){
			b = mkbin(path, lm.l_addr, err);
			if(!b){
				free(path);
				return -1;
			}
			addbin(ctl, b);
			addbs(bs, &nbs, &mbs, b);
		}
	again:
		free(path);
		lmp = lm.l_next;
	}

	return 0;
}

/* this function is a copy of scanrmap64 with 32-bit types
   for LMP and LM. */
static int
scanrmap64(Ctl *ctl, Bin ***bs, char **err)
{
	u64 lmp;
	char *path;
	struct link_map64 lm;
	Bin *b;
	unsigned nbs, mbs;
	
	nbs = mbs = 0;
	*bs = 0;
	if(0 > ctlread(ctl, ctl->as->x.r_map, &lmp, sizeof(lmp), err))
		return -1;
	while(lmp){
		if(0 > ctlread(ctl, lmp, &lm, sizeof(lm), err))
			return -1;
		path = ctlreads(ctl, lm.l_name, err);
		if(path == 0)
			return -1;
		if(strlen(path) == 0)
			goto again;
		if(!lookbinpath(ctl, path)){
			b = mkbin(path, lm.l_addr, err);
			if(!b){
				free(path);
				return -1;
			}
			addbin(ctl, b);
			addbs(bs, &nbs, &mbs, b);
		}
	again:
		free(path);
		lmp = lm.l_next;
	}

	return 0;
}

int
ldevent(Ctl *ctl, Bin ***bs, char **err)
{
	u32 x;

	if(ctl->as->x.ld == 0)
		return 0; /* static binary */
	if(0 > ctlread(ctl, ctl->as->x.r_state, &x, sizeof(x), err))
		return -1;
	switch(x){
	case RT_CONSISTENT:
		if(ctl->as->x.ldevent != RT_ADD)
			xfatal("unexpected mapping change");
		ctl->as->x.ldevent = x;
		switch(ctl->as->x.mode){
		case M32:
			if(0 > scanrmap32(ctl, bs, err))
				return -1;
			break;
		case M64:
			if(0 > scanrmap64(ctl, bs, err))
				return -1;
			break;
		default:
			bug();
		}
		return Eload;
	case RT_ADD:
		if(ctl->as->x.ldevent != RT_CONSISTENT)
			xfatal("unexpected mapping change");
		ctl->as->x.ldevent = x;
		return 0;
	case RT_DELETE:
		xfatal("unimplemented mapping change (unload)");
	default:
		xfatal("unknown mapping change");
	}

	return Eload;
}

static int
initaux(Ctl *ctl, char **err)
{
	char path[Maxpath];
	char *m;

	snprintf(path, sizeof(path), "/proc/%" PRIu64 "/auxv", ctl->id);
	m = readfile(path, 0);
	if(m == 0){
		if(err)
			*err = lasterror();
		return -1;
	}
	ctl->as->x.aux = m;
	return 0;
}


static u32
lookaux32(Ctl *ctl, unsigned long idx)
{
	u32 *u;
	u = (u32*)ctl->as->x.aux;
	while(1){
		if(u[0] == 0 && u[1] == 0)
			break;
		if(u[0] == idx)
			return u[1];
		u += 2;
	}
	return (u32)-1;
}

static u64
lookaux64(Ctl *ctl, unsigned long idx)
{
	u64 *u;
	u = (u64*)ctl->as->x.aux;
	while(1){
		if(u[0] == 0 && u[1] == 0)
			break;
		if(u[0] == idx)
			return u[1];
		u += 2;
	}
	return (u64)-1;
}

static u64
lookaux(Ctl *ctl, unsigned idx)
{
	switch(ctl->as->x.mode){
	case M32:
		return lookaux32(ctl, idx);

	case M64:
		return lookaux64(ctl, idx);
	default:
		bug();
	}
}

static char*
ctlexe(Ctl *ctl, char **err)
{
	char buf[32];
	char exe[256];
	ssize_t rv;

	snprintf(buf, sizeof(buf), "/proc/%" PRIu64 "/exe", ctl->id);
	rv = readlink(buf, exe, sizeof(exe)-1);
	if(0 > rv){
		*err = Ebin;
		return 0;
	}
	exe[rv] = 0;
	return xstrdup(exe);
}

static u64
mapaddr(Ctl *ctl, char *exepath, char **err)
{
	Stat s;
	unsigned i;

	s.pid = (int)ctl->id;
	if(0 > parsemaps(&s, err))
		return -1;
	for(i = 0; i < s.nr; i++)
		if(!strcmp(exepath, s.r[i].file))
			return s.r[i].addr;
	return -1;
}

int
syminitexec(Ctl *ctl, char **err)
{
	char *ldpath, *exepath;
	Bin *ld, *exe;
	u64 ldbase, exevbase, exeabase;
	u8 m;
	Sym *s;
	Trap t;

	exepath = ldpath = 0;
	exe = ld = 0;

	ctl->as->x.mode = execmode(ctl);
	exepath = ctlexe(ctl, err);
	if(exepath == 0)
		goto fail;
	exevbase = loadaddr(exepath, err);
	if(exevbase == (u64)-1)
		goto fail;
	exeabase = mapaddr(ctl, exepath, err);
	if(exeabase == (u64)-1)
		goto fail;
	exe = mkbin(exepath, exeabase-exevbase, err);
	if(exe == 0)
		goto fail;
	exe->isexe = 1;
	addbin(ctl, exe);

	if(0 > initaux(ctl, err))
		goto fail;

	ldpath = interp(exepath, &m, err);
	if(ldpath == 0)
		return 0; /* static executable */
	ldbase = lookaux(ctl, AT_BASE);
	if(ldbase == (u64)-1)
		goto fail;
	ld = mkbin(ldpath, ldbase, err);
	if(ld == 0)
		goto fail;
	s = looksym(ld, "_r_debug");
	if(s == 0){
		xlog("cannot find _r_debug");
		goto fail;
	}
	if(m == M32){
		ctl->as->x.r_map = s->val+sizeof(u32);
		ctl->as->x.r_state = s->val+3*sizeof(u32);
	}else{
		ctl->as->x.r_map = s->val+2*sizeof(u32);
		ctl->as->x.r_state = s->val+6*sizeof(u32);
	}

	s = looksym(ld, "_dl_debug_state");
	if(s == 0){
		xlog("cannot find _dl_debug_state");
		goto fail;
	}
	t.addr = s->val;
	t.mode = Mld;
	if(!ctlsettrap(ctl, &t, err))
		goto fail;

#if 0
	/* FIXME: inconsistent with freebsd,
	   on linux entry event is converted to load event */
	t.addr = lookaux(ctl, AT_ENTRY);
	t.mode = Mentry;
	if(!ctlsettrap(ctl, &t, err))
		goto fail;
#endif

	ctl->as->x.ld = ld;
	addbin(ctl, ld);
	free(ldpath);

	return 0;
fail:
	/* FIXME: unlink ld/exe from ctl? */
	if(ld)
		freebin(ld);
	if(exe)
		freebin(exe);
	free(exepath);
	free(ldpath);
	if(*err == 0)
		*err = Ebin;
	return -1;
}

int
syminitattach(Ctl *ctl, char **err)
{
	Bin **bs;

	if(0 > syminitexec(ctl, err))
		return -1;

	if(ctl->as->x.ld == 0)
		return 0; /* static binary */

	switch(ctl->as->x.mode){
	case M32:
		if(0 > scanrmap32(ctl, &bs, err))
			return -1;
		break;
	case M64:
		if(0 > scanrmap64(ctl, &bs, err))
			return -1;
		break;
	default:
		bug();
	}

	free(bs);
	return 0;
}

/* FIXME:
   - free break points
   - kill other ctls in same address space
 */
int
symresetexec(Ctl *ctl, char **err)
{
	symfreeas(ctl->as);
	ctl->as = mkas(ctl);
	xgetregs(ctl, &ctl->ctx.regs);	
	if(0 > syminitexec(ctl, err))
		return -1;
	xtrace(ctl, defaultevents);
	return 0;
}
