#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif
#include <fcntl.h>
#include <stdint.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/personality.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <execinfo.h>
#include <errno.h>
#include <linux/auxvec.h>

#if __WORDSIZE==32
#define SCTL32
#endif

typedef int8_t		s8;
typedef int16_t		s16;
typedef int32_t		s32;
typedef int64_t		s64;
typedef uint8_t		u8;
typedef uint16_t	u16;
typedef uint32_t	u32;
typedef uint64_t	u64;
typedef uintptr_t	uptr;

/* Decode the wait status notifications enabled by PTRACE_SETOPTIONS. */
#define WSTOPEVENT(STATUS) (((STATUS) & 0xff0000) >> 16)

#ifndef ADDR_NO_RANDOMIZE
#define ADDR_NO_RANDOMIZE	0x40000
#endif /* ADDR_NO_RANDOMIZE */

#ifndef PTRACE_SETOPTIONS
/* 0x4200-0x4300 are reserved for architecture-independent additions.  */
#define PTRACE_SETOPTIONS	0x4200
#define PTRACE_GETEVENTMSG	0x4201
#define PTRACE_GETSIGINFO	0x4202
#define PTRACE_SETSIGINFO	0x4203

/* options set using PTRACE_SETOPTIONS */
#define PTRACE_O_TRACESYSGOOD	0x00000001
#define PTRACE_O_TRACEFORK	0x00000002
#define PTRACE_O_TRACEVFORK	0x00000004
#define PTRACE_O_TRACECLONE	0x00000008
#define PTRACE_O_TRACEEXEC	0x00000010
#define PTRACE_O_TRACEVFORKDONE	0x00000020
#define PTRACE_O_TRACEEXIT	0x00000040

#define PTRACE_O_MASK		0x0000007f

/* Wait extended result codes for the above trace options.  */
#define PTRACE_EVENT_FORK	1
#define PTRACE_EVENT_VFORK	2
#define PTRACE_EVENT_CLONE	3
#define PTRACE_EVENT_EXEC	4
#define PTRACE_EVENT_VFORK_DONE	5
#define PTRACE_EVENT_EXIT	6
#endif	// PTRACE_SETOPTIONS

typedef struct {	/* Auxiliary vector entry on initial stack */
	int	a_type;			/* Entry type. */
	union {
		int	a_val;		/* Integer value. */
	} a_un;
} Elf32_Auxinfo;

typedef struct {	/* Auxiliary vector entry on initial stack */
	long	a_type;			/* Entry type. */
	union {
		long	a_val;		/* Integer value. */
		void	*a_ptr;		/* Address. */
		void	(*a_fcn)(void);	/* Function pointer (not used). */
	} a_un;
} Elf64_Auxinfo;

#define	AT_COUNT	34	/* LINUX Count of defined aux entry types. */

#define STN_UNDEF	0

// some linux installations have multiple
// definitions of user_regs_struct in system
// headers; some don't.  define it here 
// rather than navigate the mess.

#ifdef SCTL32
struct Xuser_regs_struct
{
	unsigned long ebx;
	unsigned long ecx;
	unsigned long edx;
	unsigned long esi;
	unsigned long edi;
	unsigned long ebp;
	unsigned long eax;
	unsigned long xds;
	unsigned long xes;
	unsigned long xfs;
	unsigned long xgs;
	unsigned long orig_eax;
	unsigned long eip;
	unsigned long xcs;
	unsigned long eflags;
	unsigned long esp;
	unsigned long xss;
};
#define PC eip
#define SP esp
#define DI edi
#define SI esi
#define FP ebp
#else
struct Xuser_regs_struct
{
	unsigned long r15;
	unsigned long r14;
	unsigned long r13;
	unsigned long r12;
	unsigned long rbp;
	unsigned long rbx;
	unsigned long r11;
	unsigned long r10;
	unsigned long r9;
	unsigned long r8;
	unsigned long rax;
	unsigned long rcx;
	unsigned long rdx;
	unsigned long rsi;
	unsigned long rdi;
	unsigned long orig_rax;
	unsigned long rip;
	unsigned long cs;
	unsigned long eflags;
	unsigned long rsp;
	unsigned long ss;
	unsigned long fs_base;
	unsigned long gs_base;
	unsigned long ds;
	unsigned long es;
	unsigned long fs;
	unsigned long gs;
};
#define PC rip
#define SP rsp
#define DI rdi
#define SI rsi
#define FP rbp
#endif

typedef struct Xuser_regs_struct Reg;
typedef int Fd;
typedef int Chan;
typedef struct Bin Bin;

typedef
struct Xas
{
	u32 id;			/* if non-zero, the ctl on which mem is open */
	u32 mem;		/* /proc/pid/mem fd, if id is non-zero */
	void *aux;		/* auxiliary vector */
	void *scratch;		/* breakpoint step scratch space */
	Bin *bin;		/* binaries loaded into address space */
	Bin *ld;		/* loader binary, if present */
	u64 r_map;		/* r_debug r_map address */
	u64 r_state;		/* r_debug r_state address */
	u8 ldevent;		/* last ldevent */
	u8 mode;		/* M32 or M64 */
} Xas;

typedef
struct Xctl
{
} Xctl;

typedef
struct Launch
{
	u64 flags;
	u64 narg;
	char **argv;
} Launch;

typedef
struct Thread
{
	int tid;
} Thread;

typedef
struct Proc
{
	int pid;
	char *cmd;
	u64 nth;
	Thread *th;
} Proc;

typedef
struct Ps
{
	u64 npr;
	Proc *pr;
} Ps;

typedef
struct Region
{
	char *file;
	u64 addr;
	u64 len;
	u32 flags;
} Region;

typedef
struct Stat
{
	int pid;
	u64 nr;
	Region *r;
	u64 nb;
	Bin **b;
} Stat;

typedef
struct Ctx
{
	Reg regs;
	int cursig;
} Ctx;

typedef
enum Evkind
{
	Eclear	 = 0,
	Eset	 = 1,
	Esyscall = (Eset<<1),
	Eexec	 = (Esyscall<<1),
	Efork    = (Eexec<<1),
	Eclone   = (Efork<<1),
	Esignal  = (Eclone<<1),
	Eexit    = (Esignal<<1),
	Etrap    = (Eexit<<1),
	Esnap    = (Etrap<<1),
	Estepctx = (Esnap<<1),
	Eload    = (Estepctx<<1),
	Eunload  = (Eload<<1),
} Evkind;

typedef struct Event Event;
struct Event
{
	Event *link;
	Evkind kind;
	u64 id;
	union {
		struct syscall {
			Ctx ctx;
		} syscall;
		struct trap {
			u64 id;
			Ctx ctx;
		} trap;
		struct snap {
			u64 id;
			u64 cid;
			Ctx cctx;
		} snap;
		struct spawn {
			u64 cid;
			Ctx ctx;
			Ctx cctx;
		} spawn;
		struct exec {
			Ctx ctx;
		} exec;
		struct signal {
			Ctx ctx;
			siginfo_t info;
		} signal;
		struct exit {
			Ctx ctx;
			int status;
		} exit;
		struct load {
			Bin **bs;
			Ctx ctx;
		} load;
		struct unload {
		} unload;
	} u; /* for c++ */
};
