#include "../inc/sctl.h"
#include "../inc/err.h"

typedef
struct Packed
{
	u64 r15;
	u64 r14;
	u64 r13;
	u64 r12;
	u64 rbp;
	u64 rbx;
	u64 r11;
	u64 r10;
	u64 r9;
	u64 r8;
	u64 rax;
	u64 rcx;
	u64 rdx;
	u64 rsi;
	u64 rdi;
	u64 orig_rax;
	u64 rip;
	u64 cs;
	u64 eflags;
	u64 rsp;
	u64 ss;
	u64 fs_base;
	u64 gs_base;
	u64 ds;
	u64 es;
	u64 fs;
	u64 gs;
} Packed;

u64
ctlpc(Ctl *ctl)
{
	return ctl->ctx.regs.eip;
}

void
ctlsetpc(Ctl *ctl, u64 pc)
{
	ctl->ctx.regs.eip = pc;
}

void
freectx(Ctx *c)
{
	free(c);
}

Ctx*
unpackctx(void *bytes, u64 n)
{
	Packed *p;
	Ctx *ctx;

	if(n != sizeof(Packed))
		return 0;
	p = bytes;
	ctx = xmalloc(sizeof(Ctx));
	ctx->regs.ebx = p->rbx;
	ctx->regs.ecx = p->rcx;
	ctx->regs.edx = p->rdx;
	ctx->regs.esi = p->rsi;
	ctx->regs.edi = p->rdi;
	ctx->regs.ebp = p->rbp;
	ctx->regs.eax = p->rax;
	ctx->regs.xds = p->ds;
	ctx->regs.xes = p->es;
	ctx->regs.xfs = p->fs;
	ctx->regs.xgs = p->gs;
	ctx->regs.orig_eax = p->orig_rax;
	ctx->regs.eip = p->rip;
	ctx->regs.xcs = p->cs;
	ctx->regs.eflags = p->eflags;
	ctx->regs.esp = p->rsp;
	ctx->regs.xss = p->ss;

	return ctx;
}

void
packctx(Buf *ob, Ctx *ctx)
{
	Packed p;

	memset(&p, 0, sizeof(Packed));

	p.rbx = ctx->regs.ebx;
	p.rcx = ctx->regs.ecx;
	p.rdx = ctx->regs.edx;
	p.rsi = ctx->regs.esi;
	p.rdi = ctx->regs.edi;
	p.rbp = ctx->regs.ebp;
	p.rax = ctx->regs.eax;
	p.ds = ctx->regs.xds;
	p.es = ctx->regs.xes;
	p.fs = ctx->regs.xfs;
	p.gs = ctx->regs.xgs;
	p.orig_rax = ctx->regs.orig_eax;
	p.rip = ctx->regs.eip;
	p.cs = ctx->regs.xcs;
	p.eflags = ctx->regs.eflags;
	p.rsp = ctx->regs.esp;
	p.ss = ctx->regs.xss;

	bufpush(ob, &p, sizeof(Packed));
}

u8
execmode(Ctl *ctl)
{
	return M32;
}
