#include "../inc/sctl.h"
#include "../inc/err.h"

static HT *ctltab;

int clrbrk(Ctl *ctl, Brk *bp, char **err);
int clrbrkref(Ctl *ctl, Brkref *br, char **err);

static Brkref*
mkbrkref()
{
	Brkref *br;
	br = xmalloc(sizeof(Brkref));
	return br;
}

static void
freebrkref(Brkref *br)
{
	free(br);
}

static Brk*
mkbrk()
{
	Brk *bp;
	bp = xmalloc(sizeof(Brk));
	return bp;
}

static void
freebrk(Brk *bp)
{
	free(bp);
}

static void
braddrput(HT *braddr, Brkref *br)
{
	Brkref *o;
	u64 a;
	a = br->bp->addr;
	o = hget(braddr, a);
	if(o){
		while(o->link)
			o = o->link;
		o->link = br;
		return;
	}
	hput(braddr, a, br);
}

static void
braddrdel(HT *braddr, Brkref *br)
{
	Brkref *o;
	u64 a;
	a = br->bp->addr;
	o = hget(braddr, br->bp->addr);
	if(o == br){
		hdel(braddr, br->bp->addr);
		if(o->link)
			hput(braddr, a, o->link);
	}else{
		while(o->link && o->link != br)
			o = o->link;
		if(o == 0)
			xfatal("bug");
		o->link = br->link;
	}
}

static Brkref*
braddrget(HT *braddr, u64 a)
{
	return hget(braddr, a);
}

Ctl*
_mkctl(u64 id)
{
	Ctl *ctl;
	ctl = xmalloc(sizeof(Ctl));
	ctl->id = id;
	ctl->braddr = mkht();
	ctl->brid = mkht();
	hput(ctltab, id, ctl);
	return ctl;
}

void
_freectl(Ctl *ctl)
{
	hdel(ctltab, ctl->id);
	freeht(ctl->braddr);
	freeht(ctl->brid);
	free(ctl);
}

As*
_mkas(Ctl *ctl)
{
	As *as;
	as = xmalloc(sizeof(As));
	as->ctl = ctl;
	as->bpaddr = mkht();
	return as;
}

typedef
struct Ids
{
	u32 n, max;
	u32 *ids;
} Ids;

static void
note1br(void *u, uptr k, void *v)
{
	Ids *ids;
	ids = u;
	if(ids->n >= ids->max){
		ids->ids = xrealloc(ids->ids,
				    ids->max*sizeof(u32),
				    2*ids->max*sizeof(u32));
		ids->max *= 2;
	}
	ids->ids[ids->n++] = k;
}

void
_dead(Ctl *ctl)
{
	u32 i;
	Ids ids;
	char *err;

	/* it's not safe to call ctlclrtrap in hforeach,
	   because ctlclrtrap calls hdels on the same table.
	   enumerate the keys and call it directly. */
	ctl->state = Sdead;
	ids.n = 0;
	ids.max = 16;
	ids.ids = xmalloc(ids.max*sizeof(u32));
	hforeach(ctl->brid, note1br, &ids);
	for(i = 0; i < ids.n; i++)
		/* FIXME: errors are ignored */
		ctlclrtrap(ctl, ids.ids[i], &err);
	free(ids.ids);
	ctldel(ctl->as, ctl);
	freectl(ctl);
}

static void
free1bp(void *u, uptr k, void *v)
{
	freebrk(v);
}

void
_freeas(As *as)
{
	hforeach(as->bpaddr, free1bp, 0);
	freeht(as->bpaddr);
	free(as);
}

void
ctladd(As *as, Ctl *ctl)
{
	ctl->as = as;
	ctl->link = as->ctl;
	as->ctl = ctl;
}

void
ctldel(As *as, Ctl *ctl)
{
	Ctl **p;
	p = &as->ctl;
	while(*p){
		if(*p == ctl){
			*p = ctl->link;
			if(as->ctl == 0)
				freeas(as);
			return;
		}
		p = &(*p)->link;
	}
	bug();
}

static void
brkrefcopy1(void *u, uptr k, void *v)
{
	Ctl *nctl;
	Brkref *br, *nbr;
	Brk *bp;
	nctl = u;
	br = v;
	nbr = mkbrkref();

	/* looking up bp gives us its copy
	   when we're copying an address space */
	bp = hget(nctl->as->bpaddr, br->bp->addr);
	nbr->bp = bp;

	nbr->bp->cnt++;
	nbr->id = br->id;
	hput(nctl->brid, nbr->id, nbr);
	braddrput(nctl->braddr, nbr);
}

static void
brkrefcopy(Ctl *nctl, Ctl *ctl)
{
	nctl->nextid = ctl->nextid;
	hforeach(ctl->brid, brkrefcopy1, nctl);
}

static void
bpcopy1(void *u, uptr k, void *v)
{
	As *nas;
	Brk *bp, *nbp;

	nas = u;
	bp = v;
	nbp = mkbrk();
	*nbp = *bp;
	nbp->cnt = 0;
	hput(nas->bpaddr, nbp->addr, nbp);
}

static void
bptrim1(void *u, uptr k, void *v)
{
	Ctl *ctl;
	Brk *bp;
	char *err;

	bp = v;
	if(bp->cnt > 0)
		return;
	ctl = u;
	clrbrk(ctl, bp, &err);  /* FIXME: errors are ignored */
}

void
linkspawn(Ctl *ctl, Ctl *spwn, int copy)
{
	As *nas;

	spwn->tracemode = ctl->tracemode;
	if(copy){
		nas = mkas(spwn);
		spwn->as = nas;
		hforeach(ctl->as->bpaddr, bpcopy1, nas);
		brkrefcopy(spwn, ctl);
		hforeach(spwn->as->bpaddr, bptrim1, spwn);
		symdupas(ctl->as, nas);
	}else{
		ctladd(ctl->as, spwn);
		brkrefcopy(spwn, ctl);
	}
}

Brkref*
atbrk(Ctl *ctl)
{
	Brk *bp;
	Brkref *br;

	// is this a trap for this ctl, to be handled?
	br = braddrget(ctl->braddr, ctlpc(ctl));
	if(br){
		ctl->curbrk = *br->bp;
		ctl->state = Strap;
		return br;
	}

	// is this a trap for some other ctl, to be stepped over?
	bp = hget(ctl->as->bpaddr, ctlpc(ctl));
	if(bp){
		ctl->curbrk = *bp;
		ctl->state = Strap;
		return 0;
	}

	// assume it is a recently cleared trap, to be ignored.
	return 0;
}

static void
note_clrbrk0(Ctl *ctl, Brk *bp)
{
}

int
clrbrkref(Ctl *ctl, Brkref *br, char **err)
{
	int rv;
	Brk *bp;

	rv = 0;
	bp = br->bp;
	bp->cnt--;
	if(bp->cnt == 0){
		note_clrbrk0(ctl, bp);
		rv = clrbrk(ctl, bp, err);
		hdel(ctl->as->bpaddr, bp->addr);
		freebrk(bp);
	}
	return rv;
}

static void
note_clrtrap0(Ctl *ctl, Brkref *br)
{
}

int
_ctltrace(Ctl *ctl, Trace *t, char **err)
{
	if(t->flags&Eset)
		ctl->tracemode |= t->flags;
	else
		ctl->tracemode &= ~t->flags;
	return 0;
}

int
ctlclrtrap(Ctl *ctl, u32 id, char **err)
{
	Brkref *br;
	int rv;
	u64 addr;

	rv = 0;
	br = hget(ctl->brid, id);
	if(br == 0){
		*err = Ebp;
		return -1;
	}
	addr = br->bp->addr;
	note_clrtrap0(ctl, br);
	braddrdel(ctl->braddr, br);
	rv = clrbrkref(ctl, br, err);
	hdel(ctl->brid, id);
	freebrkref(br);
	if(ctl->state != Sdead && ctl->curbrk.addr == addr)
		/* brk was hit prior to clrtrp request */
		ctl->state = Sstop;

	return rv;
}

static void
note_settrap0(Ctl *ctl, u64 addr)
{
}

static void
note_settrap1(Ctl *ctl, Brkref *br)
{
}

Brkref*
ctlsettrap(Ctl *ctl, Trap *t, char **err)
{
	Brk *bp;
	Brkref *br;

	note_settrap0(ctl, t->addr);

	bp = hget(ctl->as->bpaddr, t->addr);
	if(bp == 0){
		bp = mkbrk();
		bp->addr = t->addr;
		bp->mode = t->mode;
		if(0 > setbrk(ctl, bp, err)){
			freebrk(bp);
			return 0;
		}
		hput(ctl->as->bpaddr, bp->addr, bp);
	}else if(bp->mode != t->mode){
		/* FIXME: we should support snap and non-snap traps
		   at the same address */
		*err = Ebp;
		return 0;
	}

	br = mkbrkref();
	br->bp = bp;
	bp->cnt++;
	br->id = ctl->nextid++;
	hput(ctl->brid, br->id, br);
	braddrput(ctl->braddr, br);

	note_settrap1(ctl, br);

	return br;
}

Ctl*
lookctl(u64 id)
{
	return hget(ctltab, id);
}

char*
ctlreads(Ctl *ctl, u64 a, char **err)
{
	char *s, *p;
	unsigned m, max;

	max = 128;
	s = xmalloc(max);
	p = s;
	while(1){
		if(p-s >= max){
			m = p-s;
			s = xrealloc(s, max, max*2);
			max *= 2;
			p = s+m;
		}
		if(0 > ctlread(ctl, a, p, 1, err))
			return 0;
		if(*p == 0)
			return s;
		a++;
		p++;
	}
}

typedef
struct Ctls
{
	u32 n, max;
	Ctl **ctls;
} Ctls;

static void
note1ctl(void *u, uptr k, void *v)
{
	Ctls *ctls;
	ctls = u;
	if(ctls->n >= ctls->max){
		ctls->ctls = xrealloc(ctls->ctls,
				      ctls->max*sizeof(Ctl*),
				      2*ctls->max*sizeof(Ctl*));
		ctls->max *= 2;
	}
	ctls->ctls[ctls->n++] = v;
}

/* detach from all ctls to which we attached */
void
ctlrelease()
{
	u32 i;
	Ctls ctls;
	Ctl *ctl;
	char *err;

	/* it's not safe to call ctldetach in this hforeach
	   because it updates the ctltab */
	ctls.n = 0;
	ctls.max = 16;
	ctls.ctls = xmalloc(ctls.max*sizeof(Ctl*));
	hforeach(ctltab, note1ctl, &ctls);
	for(i = 0; i < ctls.n; i++){
		/* FIXME: errors are ignored */
		ctl = ctls.ctls[i];
		if(ctl->whence == Wattached)
			ctldetach(ctl, &err);
	}
	free(ctls.ctls);
}

/* the windows port abuses clttab */
void
ABUSE(u64 id, Ctl *ctl)
{
	hput(ctltab, id, ctl);
}

void
_sctlinit()
{
	ctltab = mkht();
}

void
_sctlfini()
{
	freeht(ctltab);
}
