#include "../inc/sctl.h"

char *op2str[] = {
/* must match enum Mkind in msg.h */
	"Reserved",
	"Rerror",
	"Reserved",
	"Aevent",
	"Tversion",
	"Rversion",
	"Tping",
	"Rping",
	"Tps",
	"Rps",
	"Tlaunch",
	"Rlaunch",
	"Tattach",
	"Rattach",
	"Tstat",
	"Rstat",
	"Tcont",
	"Rcont",
	"Tstop",
	"Rstop",
	"Tstep",
	"Rstep",
	"Tsnap",
	"Rsnap",
	"Tkill",
	"Rkill",
	"Tdetach",
	"Rdetach",
	"Ttrace",
	"Rtrace",
	"Tsettrap",
	"Rsettrap",
	"Tclrtrap",
	"Rclrtrap",
	"Tgetctx",
	"Rgetctx",
	"Tsetctx",
	"Rsetctx",
	"Tread",
	"Rread",
	"Twrite",
	"Rwrite",
	"Tlooksym",
	"Rlooksym",
	"Tenumsym",
	"Renumsym",
	"Tlooktype",
	"Rlooktype",
	"Tenumtype",
	"Renumtype",
	"Tlookaddr",
	"Rlookaddr",
	"Tenumloc",
	"Renumloc",
	"Tenumseg",
	"Renumseg",
	"Tnames",
	"Rnames",
	"Tunwind1",
	"Runwind1",
	"Tlooksrc",
	"Rlooksrc",
	"Tlookpc",
	"Rlookpc",
};

void
freemsg(Msg *m)
{
	if(m->raw)
		free(m->raw);
	free(m);
}

void
writemsg(Chan c, Msg *m)
{
	Buf *b;
	u64 n;

	xdebug(Dmessage, "%s: %s", __FUNCTION__, op2str[m->op]);

	b = mkbuf();
	bufpushu64(b, 0); 	// reserve space for message length
	bufpushu8(b, m->op);
	bufpushu32(b, m->tag);

	switch((enum Mkind)m->op){
	case Aevent:
	case Rerror:
	case Rgetctx:
	case Rps:
	case Rread:
	case Rstat:
	case Rstep:
	case Rstop:
	case Rversion:
	case Tlaunch:
	case Tping:
	case Tversion:
	case Tlooksym:
	case Rlooksym:
	case Tenumsym:
	case Renumsym:
	case Tlooktype:
	case Rlooktype:
	case Tenumtype:
	case Renumtype:
	case Tlookaddr:
	case Rlookaddr:
	case Tenumloc:
	case Renumloc:
	case Tenumseg:
	case Renumseg:
	case Tnames:
	case Rnames:
	case Tunwind1:
	case Runwind1:
	case Tlooksrc:
	case Rlooksrc:
	case Tlookpc:
	case Rlookpc:
		/* op[1] tag[4] . bytes[...] */
		bufpush(b, m->bytes, m->nbytes);
		break;
	case Ttrace:
		/* Ttrace[1] tag[4] . id[8] bytes[...] */
		bufpushu64(b, m->id);
		bufpush(b, m->bytes, m->nbytes);
		break;
	case Rattach:
	case Rlaunch:
	case Tps:
	case Tsettrap:
	case Tsetctx:
		/* op[1] tag[4] . id[8] bytes[...] */
		bufpushu64(b, m->id);
		bufpush(b, m->bytes, m->nbytes);
		break;
	case Tattach:
	case Tcont:
	case Tdetach:
	case Tgetctx:
	case Tkill:
	case Tsnap:
	case Tstat:
	case Tstep:
	case Tstop:
		/* op[1] tag[4] . id[8] */
		bufpushu64(b, m->id);
		break;
	case Rclrtrap:
	case Rcont:
	case Rdetach:
	case Rkill:
	case Rsetctx:
	case Rtrace:
		/* op[1] tag[4] . */
		break;
	case Rsettrap:
		/* op[1] tag[4] . tid[8] */
		bufpushu64(b, m->tid);
		break;
	case Tclrtrap:
		/* op[1] tag[4] . id[8] tid[8] */
		bufpushu64(b, m->id);
		bufpushu64(b, m->tid);
		break;
	case Rsnap:
		/* Rsnap[1] tag[4] . cid[8] bytes[...] */
		bufpushu64(b, m->cid);
		bufpush(b, m->bytes, m->nbytes);
		break;
	case Tread:
		/* Tread[1] tag[4] id[8] . fd[8] offset[8] cnt[8] */
		bufpushu64(b, m->id);
		bufpushu64(b, m->fd);
		bufpushu64(b, m->addr);
		bufpushu64(b, m->cnt);
		break;
	case Twrite:
		/* Twrite[1] tag[4] id[8] . fd[8] offset[8] bytes[...] */
		bufpushu64(b, m->id);
		bufpushu64(b, m->fd);
		bufpushu64(b, m->addr);
		bufpush(b, m->bytes, m->nbytes);
		break;
	case Rping:
	case Rwrite:
		/* op[1] tag[4] . cnt[8] */
		bufpushu64(b, m->cnt);
		break;
	default:
		xfatal("bug");
	}

	n = b->len-sizeof(u64);
	bufset(b, 0, &n, sizeof(u64));
	xsend(c, b->buf, b->len);
	freebuf(b);
}

#undef XRD
#define XRD(to,fr,sz) { 		\
	if(len < (sz)) goto fail;	\
	memcpy((to),(fr),(sz));		\
	(fr) += (sz);			\
	len -= (sz);			\
}

Msg*
readmsg(Chan c)
{
	Msg *m;
	char *p;
	u64 len;
	int rv;

	rv = xrecv(c, &len, sizeof(u64));
	if(0 > rv || rv != sizeof(u64))
		return 0;
	m = xmalloc(sizeof(Msg));
	m->raw = xmalloc(len);		/* freed by freemsg */
	rv = xrecv(c, m->raw, len);
	if(0 > rv || rv != len)
		goto fail;
	if(len < sizeof(u8)+sizeof(u32))
		/* every message begins with op[1] tag[4] */
		goto fail;
	p = m->raw;
	XRD(&m->op, p, sizeof(u8));
	XRD(&m->tag, p, sizeof(u32));

	xdebug(Dmessage, "%s: %s", __FUNCTION__, op2str[m->op]);
	switch((enum Mkind)m->op){
	case Rerror:
	case Rps:
	case Rstat:
	case Rversion:
	case Tlaunch:
	case Tping:
	case Tps:
	case Tversion:
	case Tlooksym:
	case Rlooksym:
	case Tenumsym:
	case Renumsym:
	case Tlooktype:
	case Rlooktype:
	case Tenumtype:
	case Renumtype:
	case Tlookaddr:
	case Rlookaddr:
	case Tenumloc:
	case Renumloc:
	case Tenumseg:
	case Renumseg:
	case Tnames:
	case Rnames:
	case Tunwind1:
	case Runwind1:
	case Tlooksrc:
	case Rlooksrc:
	case Tlookpc:
	case Rlookpc:
		/* op[1] tag[4] . bytes[...] */
		m->nbytes = len;
		m->bytes = p;
		break;
	case Rattach:
	case Rlaunch:
	case Rsnap:
	case Tsetctx:
	case Tsettrap:
		/* op[1] tag[4] . id[8] bytes[...] */
		XRD(&m->id, p, sizeof(m->id));
		m->nbytes = len;
		m->bytes = p;
		break;
	case Ttrace:
		/* op[1] tag[4] . id[8] bytes[...] */
		XRD(&m->id, p, sizeof(m->id));
		m->nbytes = len;
		m->bytes = p;
		break;
	case Tattach:
	case Tcont:
	case Tdetach:
	case Tgetctx:
	case Tkill:
	case Tstop:
	case Tsnap:
	case Tstat:
	case Tstep:
		/* op[1] tag[4] . id[8] */
		XRD(&m->id, p, sizeof(m->id));
		break;
	case Rgetctx:
	case Rread:
	case Rstep:
	case Rstop:
		/* op[1] tag[4] . data[..] */
		m->nbytes = len;
		m->bytes = p;
		break;
	case Rclrtrap:
	case Rcont:
	case Rdetach:
	case Rkill:
	case Rsetctx:
	case Rtrace:
		/* op[1] tag[4] . */
		break;
	case Rsettrap:
		/* Rsettrap[1] tag[4] . tid[8] */
		XRD(&m->tid, p, sizeof(m->tid));
		break;
	case Tclrtrap:
		/* Tclrtrap[1] tag[4] . id[8] tid[8] */
		XRD(&m->id, p, sizeof(m->id));
		XRD(&m->tid, p, sizeof(m->tid));
		break;
	case Aevent:
		/* Aevent[1] tag[4] . id[8] bytes[...] */
		XRD(&m->id, p, sizeof(m->id));
		m->nbytes = len;
		m->bytes = p;
		break;
	case Tread:
		/* Tread[1] tag[4] . id[8] fd[8] addr[8] cnt[8] */
		XRD(&m->id, p, sizeof(m->id));
		XRD(&m->fd, p, sizeof(m->fd));
		XRD(&m->addr, p, sizeof(m->addr));
		XRD(&m->cnt, p, sizeof(m->cnt));
		break;
	case Twrite:
		/* Twrite[1] tag[4] . id[8] fd[8] addr[8] bytes[...] */
		XRD(&m->id, p, sizeof(m->id));
		XRD(&m->fd, p, sizeof(m->fd));
		XRD(&m->addr, p, sizeof(m->addr));
		m->nbytes = len;
		m->bytes = p;
		break;
	case Rping:
	case Rwrite:
		/* op[1] tag[4] . cnt[8] */
		XRD(&m->cnt, p, sizeof(m->cnt));
		break;
	}

	return m;
fail:
	freemsg(m);
	return 0;
}
