#include "../inc/sctl.h"
#include "../inc/err.h"

static char prversion[] = "sctl-2012";
static char *curversion = 0;
enum Mkind notused; 	/* force older gcc to define Mkind in dwarf */

void
closechan(Chan c)
{
	free(curversion);
}

void
rerror(Chan c, u32 tag, char *err)
{
	Msg m;
	m.op = Rerror;
	m.tag = tag;
	m.bytes = err;
	m.nbytes = strlen(err);
	writemsg(c, &m);
}

void
aevent(Chan c, Event *e)
{
	Msg rep;
	Buf *b;
	if(e == 0)
		return;
	rep.op = Aevent;
	rep.tag = 1;
	b = mkbuf();
	packevent(b, e);
	rep.bytes = b->buf;
	rep.nbytes = b->len;
	writemsg(c, &rep);
	freebuf(b);
	aevent(c, e->link);
}

#undef XRD
#define XRD(to,fr,sz) { 		\
	if(len < (sz)) goto fail;	\
	memcpy((to),(fr),(sz));		\
	(fr) += (sz);			\
	len -= (sz);			\
}

void
freetrap(Trap *t)
{
	free(t);
}

Trap*
unpacktrap(void *buf, u64 len)
{
	char *p;
	Trap *t;

	t = xmalloc(sizeof(Trap));
	p = buf;
	XRD(&t->mode, p, sizeof(t->mode));
	XRD(&t->addr, p, sizeof(t->addr));
	return t;
fail:
	free(t);
	return 0;
}

void
freetrace(Trace *t)
{
	free(t);
}

Trace*
unpacktrace(void *buf, u64 len)
{
	char *p;
	Trace *t;

	t = xmalloc(sizeof(Trace));
	p = buf;
	XRD(&t->flags, p, sizeof(t->flags));
	return t;
fail:
	free(t);
	return 0;
}

static char*
negotiate(char *v, u64 len)
{
	char *pr, *os, *rv;
	unsigned m;

	rv = 0;
	pr = xmalloc(len+1);
	memcpy(pr, v, len); /* nul-terminated */
	os = strchr(pr, ':');
	if(os == 0)
		goto out;
	*os++ = 0;
	if(strcmp(pr, prversion) != 0)
		goto out;
	if(strstr(os, osversion) == 0)
		goto out;
	m = strlen(prversion)+1+strlen(osversion)+1;
	rv = xmalloc(m);
	snprintf(rv, m, "%s:%s", prversion, osversion);
out:
	free(pr);
	return rv;
}

void
domsg(Chan c, Msg *m)
{
	Msg rep;
	Brkref *br;
	Ctl *ctl, *spwn;
	char *err;
	Buf *xbuf;
	Ctx *ctx;
	Launch *l;
	Ps *p;
	Stat *s;
	Trap *tp;
	Trace *tc;

	if(curversion == 0 && m->op != Tversion){
		rerror(c, m->tag, Eproto);
		return;
	}

	err = 0;
	xdebug(Daction, "%s(%s)", __FUNCTION__, op2str[m->op]);
	switch(m->op){
	case Tversion:
		if(curversion){
			rerror(c, m->tag, Eproto);
			return;
		}
		curversion = negotiate(m->bytes, m->nbytes);
		if(curversion == 0){
			rerror(c, m->tag, Eversion);
			return;
		}
		rep.op = Rversion;
		rep.tag = m->tag;
		rep.nbytes = strlen(curversion);
		rep.bytes = curversion;
		writemsg(c, &rep);
		break;
	case Tlaunch:
		l = unpacklaunch(m->bytes, m->nbytes);
		if(l == 0){
			rerror(c, m->tag, Eproto);
			return;
		}
		ctl = launch(l, &err);
		freelaunch(l);
		if(ctl == 0){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rlaunch;
		rep.tag = m->tag;
		rep.id = ctl->id;
		xbuf = mkbuf();
		packctx(xbuf, &ctl->ctx);
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	case Tattach:
		ctl = attach(m->id, &err);
		if(ctl == 0){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rattach;
		rep.tag = m->tag;
		rep.id = ctl->id;
		xbuf = mkbuf();
		packctx(xbuf, &ctl->ctx);
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	case Ttrace:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		tc = unpacktrace(m->bytes, m->nbytes);
		if(tc == 0){
			rerror(c, m->tag, Eproto);
			return;
		}
		switch(ctl->state){
		case Sdead:
			freetrace(tc);
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		if(0 > ctltrace(ctl, tc, &err)){
			freetrace(tc);
			rerror(c, m->tag, err);
			return;
		}
		freetrace(tc);
		rep.op = Rtrace;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tcont:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Strap:
			if(0 > steptrap(ctl, &err)){
				rerror(c, m->tag, err);
				return;
			}
			/* fall through */
		case Sstop:
			ctlcont(ctl);
			break;
		case Srun:
			break;
		default:
			bug();
		}
		rep.op = Rcont;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tstop:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			ctlstop(ctl, &err);
			if(err){
				rerror(c, m->tag, err);
				return;
			}
			break;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		rep.op = Rstop;
		rep.tag = m->tag;
		xbuf = mkbuf();
		packctx(xbuf, &ctl->ctx);
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	case Tdetach:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Strap:
		case Sstop:
			if(0 > ctldetach(ctl, &err)){
				rerror(c, m->tag, err);
				return;
			}
			break;
		default:
			bug();
		}
		rep.op = Rdetach;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tkill:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Strap:
		case Sstop:
			if(0 > ctlkill(ctl, &err)){
				rerror(c, m->tag, err);
				return;
			}
			break;
		default:
			bug();
		}
		rep.op = Rkill;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tsnap:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		spwn = ctlsnap(ctl);
		rep.op = Rsnap;
		rep.tag = m->tag;
		rep.cid = spwn->id;
		xbuf = mkbuf();
		packctx(xbuf, &spwn->ctx);
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	case Tsettrap:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		tp = unpacktrap(m->bytes, m->nbytes);
		if(tp == 0){
			rerror(c, m->tag, Eproto);
			return;
		}
		switch(ctl->state){
		case Sdead:
			freetrap(tp);
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		br = ctlsettrap(ctl, tp, &err);
		freetrap(tp);
		if(br == 0){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rsettrap;
		rep.tag = m->tag;
		rep.tid = br->id;
		writemsg(c, &rep);
		break;
	case Tread:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		rep.bytes = xmalloc(m->cnt);
		rep.nbytes = m->cnt;
		if(0 > ctlread(ctl, m->addr, rep.bytes, m->cnt, &err)){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rread;
		rep.tag = m->tag;
		writemsg(c, &rep);
		free(rep.bytes);
		break;
	case Twrite:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		if(0 > ctlwrite(ctl, m->addr, m->bytes, m->nbytes, &err)){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rwrite;
		rep.tag = m->tag;
		rep.cnt = m->nbytes;
		writemsg(c, &rep);
		break;
	case Tclrtrap:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		if(0 > ctlclrtrap(ctl, m->tid, &err)){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rclrtrap;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tgetctx:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		rep.op = Rgetctx;
		rep.tag = m->tag;
		ctx = ctlgetctx(ctl, &err);
		if(ctx == 0){
			rerror(c, m->tag, err);
			return;
		}
		xbuf = mkbuf();
		packctx(xbuf, ctx);
		freectx(ctx);
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	case Tsetctx:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}
		ctx = unpackctx(m->bytes, m->nbytes);
		if(ctx == 0){
			rerror(c, m->tag, Eproto);
			return;
		}
		if(0 > ctlsetctx(ctl, ctx, &err)){
			freectx(ctx);
			rerror(c, m->tag, err);
			return;
		}
		freectx(ctx);
		rep.op = Rsetctx;
		rep.tag = m->tag;
		writemsg(c, &rep);
		break;
	case Tps:
		p = ps(&err);
		if(p == 0){
			rerror(c, m->tag, err);
			return;
		}
		xbuf = mkbuf();
		packps(xbuf, p);
		rep.op = Rps;
		rep.tag = m->tag;
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freeps(p);
		freebuf(xbuf);
		break;
	case Tstat:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
		case Strap:
			break;
		default:
			bug();
		}

		s = ctlstat(ctl, &err);
		if(s == 0){
			rerror(c, m->tag, err);
			return;
		}
		xbuf = mkbuf();
		packstat(xbuf, s);
		rep.op = Rstat;
		rep.tag = m->tag;
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		freestat(s);
		break;
	case Tping:
		rep.op = Rping;
		rep.tag = m->tag;
		rep.cnt = m->nbytes;
		writemsg(c, &rep);
		break;
	case Tstep:
		ctl = lookctl(m->id);
		if(ctl == 0){
			rerror(c, m->tag, Eid);
			return;
		}
		switch(ctl->state){
		case Sdead:
			rerror(c, m->tag, Eid);
			return;
		case Srun:
		case Strap:
			rerror(c, m->tag, Eready);
			return;
		case Sstop:
			break;
		default:
			bug();
		}
		if(0 > ctlstep(ctl, &err)){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = Rstep;
		rep.tag = m->tag;
		rep.nbytes = 0;
		if(ctl->tracemode&Estepctx){
			xbuf = mkbuf();
			packctx(xbuf, &ctl->ctx);
			rep.bytes = xbuf->buf;
			rep.nbytes = xbuf->len;
			writemsg(c, &rep);
			freebuf(xbuf);
		}else
			writemsg(c, &rep);
		break;
	case Tlooksym:
	case Tenumsym:
	case Tlooktype:
	case Tenumtype:
	case Tlookaddr:
	case Tenumloc:
	case Tenumseg:
	case Tnames:
	case Tunwind1:
	case Tlooksrc:
	case Tlookpc:
		xbuf = mkbuf();
		if(0 > ctlns(m->op, m->bytes, m->nbytes, xbuf, &err)){
			rerror(c, m->tag, err);
			return;
		}
		rep.op = m->op+1;
		rep.tag = m->tag;
		rep.bytes = xbuf->buf;
		rep.nbytes = xbuf->len;
		writemsg(c, &rep);
		freebuf(xbuf);
		break;
	default:
		rerror(c, m->tag, Eproto);
		break;
	}
}
