#include "../inc/sctl.h"

static char* cbasename[] = {
	[Vundef]              = "error!",
	[Vbool]		      = "_Bool",
	[Vchar]               = "char",
	[Vshort]	      = "short",
	[Vint]		      = "int",
	[Vlong]		      = "long",
	[Vvlong]	      = "long long",
	[Vuchar]	      = "unsigned char",
	[Vushort]	      = "unsigned short",
	[Vuint]		      = "unsigned int",
	[Vulong]	      = "unsigned long",
	[Vuvlong]	      = "unsigned long long",
	[Vfloat]	      = "float",
	[Vdouble]	      = "double",
	[Vlongdouble]	      = "long double",
};

char* tkindstr[] = {
	[Tarr]			= "array",
	[Tbase]			= "base",
	[Tenum]			= "enum",
	[Tfun]			= "function",
	[Tptr]			= "pointer",
	[Tstruct]		= "struct",
	[Ttypedef]		= "typedef",
	[Tunion]		= "union",
	[Tvoid]			= "void",
	[Tbitfield]		= "bitfield",
	[Tconst]		= "constant",
	[Tundef]		= "undef",
};

static char*	_fmttype(Type *t, char *o);

Enum*
mkenumconst()
{
	Enum *e;
	e = xmalloc(sizeof(Enum));
	return e;
}

void
freeenumconst(Enum *e)
{
	free(e->id);
	/* don't free the type */
	free(e);
}

Decl*
mkdecl()
{
	Decl *d;
	d = xmalloc(sizeof(Decl));
	return d;
}

void
freedecl(Decl *d)
{
	if(d == 0)
		return;
	free(d->id);
	freetype(d->type);
	free(d);
}

Type*
mktype()
{
	Type *t;
	t = xmalloc(sizeof(Type));
	return t;
}

Type*
mkvoid()
{
	Type *t;
	t = mktype();
	t->kind = Tbase;
	t->base = Vvoid;
	t->rep = Rundef;
	return t;
}

Type*
mksue(Tkind kind, char *tag)
{
	Type *t;
	t = mktype();
	t->kind = kind;
	t->id = xstrdup(tag);
	return t;
}

Type*
mkstruct(char *tag)
{
	return mksue(Tstruct, tag);
}

Type*
mkunion(char *tag)
{
	return mksue(Tunion, tag);
}

Type*
mkenum(char *tag)
{
	return mksue(Tenum, tag);
}

Type*
mkptr(Type *st)
{
	Type *t;
	t = mktype();
	t->kind = Tptr;
	t->link = st;
	return t;
}

Type*
mktypedef(char *tid)
{
	Type *t;
	t = mktype();
	t->kind = Ttypedef;
	t->id = xstrdup(tid);
	return t;
}

void
freetype(Type *t)
{
	u32 i;
	if(t == 0)
		return;
	switch(t->kind){
	case Tbase:
		break;
	case Tstruct:
	case Tunion:
		free(t->id);
		if(t->ds)
			for(i = 0; i < t->n; i++)
				freedecl(t->ds[i]);
		free(t->ds);
		break;
	case Tenum:
		free(t->id);
		if(t->en)
			for(i = 0; i < t->n; i++)
				freeenumconst(t->en[i]);
		free(t->en);
		break;
	case Ttypedef:
		free(t->id);
		freetype(t->link);
		break;
	case Tptr:
	case Tarr:
	case Tconst:
	case Tbitfield:
		freetype(t->link);
		break;
	case Tfun:
		if(t->ds){
			for(i = 0; i < t->n; i++)
				freedecl(t->ds[i]);
			free(t->ds);
		}
		freetype(t->link);
		break;
	default:
		xfatal("unexpected type %d", t->kind);
	}
	free(t);
}

Enum*
copyenum(Enum *e, Type *t)
{
	Enum *u;
	u = mkenumconst();
	*u = *e;
	u->id = xstrdup(e->id);
	u->type = t;
	return u;
}

Decl*
copydecl(Decl *d)
{
	Decl *u;
	u = mkdecl();
	*u = *d;
	u->id = xstrdup(d->id);
	u->type = copytype(d->type);
	return u;
}

Type*
copytype(Type *t)
{
	Type *u;
	u32 i;

	if(t == 0)
		return 0;
	u = mktype();
	*u = *t;
	switch(t->kind){
	case Tbase:
		break;
	case Tstruct:
	case Tunion:
		u->id = xstrdup(t->id);
		if(t->ds){
			u->ds = xmalloc(t->n*sizeof(Decl*));
			for(i = 0; i < t->n; i++)
				u->ds[i] = copydecl(t->ds[i]);
		}
		break;
	case Tenum:
		u->id = xstrdup(t->id);
		if(t->en){
			u->en = xmalloc(t->n*sizeof(Enum*));
			for(i = 0; i < t->n; i++)
				u->en[i] = copyenum(t->en[i], u);
		}
		break;
	case Tptr:
	case Tarr:
		u->link = copytype(t->link);
		break;
	case Ttypedef:
		u->id = xstrdup(t->id);
		u->link = copytype(t->link);
		break;
	case Tfun:
		if(t->ds){
			u->ds = xmalloc(t->n*sizeof(Decl*));
			for(i = 0; i < t->n; i++)
				u->ds[i] = copydecl(t->ds[i]);
		}
		u->link = copytype(t->link);
		break;
	default:
		bug();
	}
	return u;
}

Types*
mktypes()
{
	Types *ts;
	ts = xmalloc(sizeof(Types));
	ts->max = 16;
	ts->t = xmalloc(ts->max*sizeof(Type*));
	return ts;
}

void
addtype(Types *ts, Type *t)
{
	if(ts->n >= ts->max){
		ts->t = xrealloc(ts->t, ts->max*sizeof(Type*),
				 2*ts->max*sizeof(Type*));
		ts->max *= 2;
	}
	ts->t[ts->n++] = t;
}

void
freetypes(Types *ts)
{
	u32 i;
	for(i = 0; i < ts->n; i++)
		freetype(ts->t[i]);
	free(ts->t);
	free(ts);
}

static char*
fmtplist(Decl **param, u32 n)
{
	char **ds, *buf, *bp, *id;
	u32 i, m;
	Type *t;
	Decl *d;

	if(n == 0)
		return xstrdup("");

	ds = xmalloc(n*sizeof(char**));
	m = 1;			/* null */
	for(i = 0; i < n; i++){
		d = param[i];
		t = d->type;
		id = d->id;
		if(id)
			ds[i] = fmtdecl(t, id);
		else
			ds[i] = _fmttype(t, xstrdup(""));
		m += strlen(ds[i]);
		if(i < n-1)
			m += 2;	/* comma, space */
	}
	buf = xmalloc(m);
	bp = buf;
	for(i = 0; i < n; i++){
		strcpy(bp, ds[i]);
		bp += strlen(ds[i]);
		free(ds[i]);
		if(i < n-1){
			strcpy(bp, ", ");
			bp += 2;
		}
	}
	*bp = 0;
	free(ds);
	return buf;
}

static char*
_fmttype(Type *t, char *o)
{
	char *buf, *w, *pl, *id, *b;
	u32 m, leno;
	Type *st;

	leno = strlen(o);
	switch(t->kind){
	case Tvoid:
		m = 4+1+leno+1;
		buf = xmalloc(m);
		if(leno)
			snprintf(buf, m, "void %s", o);
		else
			snprintf(buf, m, "void");
		free(o);
		return buf;
	case Tbase:
		b = cbasename[t->base];
		m = strlen(b)+1+leno+1;
		buf = xmalloc(m);
		if(leno)
			snprintf(buf, m, "%s %s", b, o);
		else
			snprintf(buf, m, "%s", b);
		free(o);
		return buf;
	case Ttypedef:
		id = t->id;
		m = strlen(id)+1+leno+1;
		buf = xmalloc(m);
		if(leno)
			snprintf(buf, m, "%s %s", id, o);
		else
			snprintf(buf, m, "%s", id);
		free(o);
		return buf;
	case Tstruct:
	case Tunion:
	case Tenum:
		w = tkindstr[t->kind];
		id = t->id;
		if(id){
			m = strlen(w)+1+strlen(id)+1+leno+1;
			buf = xmalloc(m);
			if(leno)
				snprintf(buf, m, "%s %s %s", w, id, o);
			else
				snprintf(buf, m, "%s %s", w, id);
		}else{
			m = strlen(w)+1+leno+1;
			buf = xmalloc(m);
			if(leno)
				snprintf(buf, m, "%s %s", w, o);
			else
				snprintf(buf, m, "%s", w);
		}
		free(o);
		return buf;
	case Tptr:
		m = 2+leno+1+1;
		buf = xmalloc(m);
		st = t->link;
		if(st->kind == Tfun || st->kind == Tarr)
			snprintf(buf, m, "(*%s)", o);
		else
			snprintf(buf, m, "*%s", o);
		free(o);
		return _fmttype(st, buf);
	case Tarr:
		m = leno+1+10+1+1;	/* assume max 10 digit size */
		buf = xmalloc(m);
		if(t->n == 0)
			snprintf(buf, m, "%s[]", o);
		else
			snprintf(buf, m, "%s[%" PRIu32 "]", o, t->n);
		free(o);
		return _fmttype(t->link, buf);
	case Tfun:
		pl = fmtplist(t->ds, t->n);
		m = leno+1+strlen(pl)+1+1;
		buf = xmalloc(m);
		snprintf(buf, m, "%s(%s)", o, pl);
		free(o);
		free(pl);
		return _fmttype(t->link, buf);
	case Tundef:
	case Tconst:
	case Tbitfield:
		/* we may want a more accurate printed
		   representation of these objects,
		   especially bitfields */
		return _fmttype(t->link, o);
	default:
		bug();
	}
	return 0;

}

char*
fmtdecl(Type *t, char *id)
{
	char *o;
	u32 m;
	m = strlen(id);
	o = xmalloc(m+1);
	memcpy(o, id, m);
	return _fmttype(t, o);
}

char*
fmttype(Type *t)
{
	return _fmttype(t, xstrdup(""));
}
