#include "../inc/sctl.h"

/* options */
int debuglevel = 0;
int ignoreg = 0;

char *argv0;

void
setargv0(char *p)
{
	argv0 = p;
}

void
xlog(char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "%s: ", argv0);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	va_end(args);
}

void
xdebug(int level, char *fmt, ...)
{
	va_list args;

	if(debuglevel < level)
		return;
	fprintf(stderr, "[%d] ", getpid());
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr, "\n");
}

void 
xdebughex(int level, char *msg, char *buf, unsigned len)
{
	unsigned i;
	if(debuglevel < level)
		return;
	fprintf(stderr, "[%d] %s (%d):", getpid(), msg, len);
	for (i = 0; i < len; i++)
		fprintf(stderr, "%.2x ", buf[i]);
	fprintf(stderr, "\n");
}

void
xfatal(char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "%s: ", argv0);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	va_end(args);
	dumpstack();
	killall();
}

void
xfatalno(char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "%s: ", argv0);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, ": %s\n", lasterror());
	va_end(args);
	killall();
}

char *
xstrdup(char *s)
{
	s = strdup(s);
	if(s == 0)
		xfatal("out of memory");
	return s;
}

void *
xmalloc(size_t size)
{
	void *p;
	p = malloc(size);
	if(p == 0)
		xfatal("out of memory");
	memset(p, 0, size);
	return p;
}

void*
xrealloc(void *p, size_t osz, size_t sz)
{
	p = realloc(p, sz);
	if(p == 0)
		xfatal("out of memory");
	if(sz > osz)
		memset((char *)p+(unsigned long)osz, 0, sz-osz);
	return p;
}

void
nodelay(Chan c)
{
	int optval = 1;
	if (0 > setsockopt(c, IPPROTO_TCP, TCP_NODELAY,
			   (const char *)&optval, sizeof(optval)))
		xfatalno("nodelay");
}

void
reuseaddr(Chan c)
{
	int optval = 1;
	if (0 > setsockopt(c, SOL_SOCKET, SO_REUSEADDR,
			   (const char *)&optval, sizeof(optval)))
		xfatalno("reuseaddr");
}
