typedef
struct Attrspec
{
	u32 name, form;
} Attrspec;

typedef
struct Abbrev
{
	u32 code, tag, haskids;
	u32 nattr;
	Attrspec *attr;
} Abbrev;

typedef
struct Abbrevtab
{
	u32 nab;
	Abbrev *ab;
} Abbrevtab;

typedef
struct Attr
{
	u32 name, form;
	u64  x64;
	s64 sx64;
	char *str;
	char *block;
} Attr;

typedef
struct Srcloc
{
	u64 addr;
	char *file;
	s32 line, col;
} Srcloc;

typedef
struct Src
{
	char **file;		/* storage for file names */
	u32 nfile;
	Srcloc *loc;
	u32 n, max;
} Src;

typedef struct Die Die;
typedef struct CU CU;
struct Die
{
	u32 ab;
	u32 tag;
	u32 nattr;
	Attr *attr;
	Die *kid;
	Die *sib;
	CU *cu;
	char *id;		/* resolved anonymous struct/union/enum's */
};

struct CU
{
	Die *d;			/* root of Die tree */
	u32 sz;			/* address size in bytes specified in header */
	u32 off;		/* offset in debuginfo */
	char *p;		/* raw cu in debuginfo */
	HT *idx;		/* cu-relative offsets to dies */
	Src *src;		/* source line information */
	char *dir;		/* compilation unit directory */
	char *file;		/* compilation unit file */
	u64 soff;		/* statement list offset */
};

typedef
struct Unwinder
{
	char *cie;		/* address of cie */
	char *f, *ef;		/* start/end of fde instructions */
	char *c, *ec;		/* start/end of cie instructions */
	u64 lo, hi;		/* fde pc range */
	u64 caf;		/* code alignment factor */
	s64 daf;		/* data alignment factor */
	u64 rar;		/* return address register */
	u8 z, enc;		/* augmentation */
} Unwinder;

typedef
struct Diet
{
	u32 cuoff;		/* compile unit offset */
	u32 dieoff;		/* cu-relative die offset */
} Diet;

typedef
struct Diets
{
	Diet **diet;
	u32 n, max;
} Diets;

typedef
struct Alist
{
	int (*eq)(void *k1, void *k2);
	void **k, **v;
	u32 n, max;
} Alist;

void		bufpushsrc(Buf *x, Srcloc *l);
void		freesrc(Src *src);
int		lookpc(Bin *b, char *file, u32 line, u64 *addr, char **err);
Srcloc*		looksrc(Bin *b, u64 pc, char **err);

int		lookarange(Bin *b, u64 addr, u32 *op);
Attr*		lookattr(Die *d, u32 n);
CU*		lookcu(Bin *b, u32 cuoff);
s64		sleb32(char **p);
u64		uleb(char **p);
