#include "../inc/sctl.h"
#include "../inc/err.h"
#include "unix.h"

void
freeevent(Event *e)
{
	if(e == 0)
		return;
	freeevent(e->link);
	if(e->kind == Eload)
		free(e->u.load.bs);
	free(e);
}

Event*
eventsyscall(Ctl *ctl)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Esyscall;
	e->u.syscall.ctx = ctl->ctx;
	return e;
}

Event*
eventtrap(Ctl *ctl, Brkref *br)
{
	Event *e;
	if(br == 0)
		return 0;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Etrap;
	e->u.trap.id = br->id;
	e->u.trap.ctx = ctl->ctx;
	e->link = eventtrap(ctl, br->link);
	return e;
}

Event*
eventsnap(Ctl *ctl, Brkref *br, Ctl *spwn)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Esnap;
	e->u.snap.id = br->id;
	e->u.snap.cid = spwn->id;
	e->u.snap.cctx = spwn->ctx;
	return e;
}

Event*
eventspawn(Evkind kind, Ctl *ctl,  Ctl *spwn)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = kind;
	e->u.spawn.cid = spwn->id;
	e->u.spawn.ctx = ctl->ctx;
	e->u.spawn.cctx = spwn->ctx;
	return e;
}

Event*
eventclone(Ctl *ctl, Ctl *spwn)
{
	return eventspawn(Eclone, ctl, spwn);
}

Event*
eventfork(Ctl *ctl, Ctl *spwn)
{
	return eventspawn(Efork, ctl, spwn);
}

Event*
eventexit(Ctl *ctl, int st)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eexit;
	e->u.exit.status = st;
	e->u.exit.ctx = ctl->ctx;
	return e;
}

Event*
eventexec(Ctl *ctl)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eexec;
	e->u.exec.ctx = ctl->ctx;
	return e;
}

Event*
eventsignal(Ctl *ctl, siginfo_t *info)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Esignal;
	e->u.signal.ctx = ctl->ctx;
	memcpy(&e->u.signal.info, info, sizeof(e->u.signal.info));
	return e;
}

Event*
eventload(Ctl *ctl, Bin **bs)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eload;
	e->u.load.bs = bs;
	e->u.load.ctx = ctl->ctx;
	return e;
}

static void
packload(Buf *b, Event *e)
{
	u64 m;
	Bin **bs;
	bs = e->u.load.bs;
	if(bs == 0)
		xfatal("bug");
	m = 0;
	while(*bs++)
		m++;
	bs = e->u.load.bs;
	bufpushu64(b, m);
	while(m-- > 0){
		bufpushs(b, (*bs)->path);
		bufpushu64(b, (*bs)->base);
		bs++;
	}
	packctx(b, &e->u.load.ctx);
}

void
packevent(Buf *b, Event *e)
{
	bufpushu64(b, e->id);
	bufpushu64(b, e->kind);
	switch(e->kind){
	case Esyscall:
		packctx(b, &e->u.syscall.ctx);
		break;
	case Etrap:
		bufpushu64(b, e->u.trap.id);
		packctx(b, &e->u.trap.ctx);
		break;
	case Esnap:
		bufpushu64(b, e->u.snap.id);
		bufpushu64(b, e->u.snap.cid);
		packctx(b, &e->u.snap.cctx);
		break;
	case Eclone:
	case Efork:
		bufpushu64(b, e->u.spawn.cid);
		packctx(b, &e->u.spawn.ctx);
		packctx(b, &e->u.spawn.cctx);
		break;
	case Eexit:
		bufpushu64(b, e->u.exit.status);
		packctx(b, &e->u.exit.ctx);
		break;
	case Eexec:
		packctx(b, &e->u.exec.ctx);
		break;
	case Esignal:
		packsiginfo(b, &e->u.signal.info);
		packctx(b, &e->u.signal.ctx);
		break;
	case Eload:
	case Eunload:
		packload(b, e);
		break;
	default:
		bug();
	}
}
