#include "../inc/sctl.h"
#include "ns.h"

char* xtypename[Xn] = {
	[Xptr]=		"uintptr",
	[Xint8]=	"int8",
	[Xint16]=	"int16",
	[Xint32]=	"int32",
	[Xint64]=	"int64",
	[Xuint8]=	"uint8",
	[Xuint16]=	"uint16",
	[Xuint32]=	"uint32",
	[Xuint64]=	"uint64",
};

Rootns rootns[Rnr] = {
[Rc32le] = {
	.name = "c32le",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08le,
		[Vshort]=	Rs16le,
		[Vint]=		Rs32le,
		[Vlong]=	Rs32le,
		[Vvlong]=	Rs64le,
		[Vuchar]=	Ru08le,
		[Vushort]=	Ru16le,
		[Vuint]=	Ru32le,
		[Vulong]=	Ru32le,
		[Vuvlong]=	Ru64le,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru32le,
	},
	.xtype = {
		[Xptr]=		Vulong,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vlong,
		[Xint64]=	Vvlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vulong,
		[Xuint64]=	Vuvlong,
	},
},
[Rc32be] = {
	.name = "c32be",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08be,
		[Vshort]=	Rs16be,
		[Vint]=		Rs32be,
		[Vlong]=	Rs32be,
		[Vvlong]=	Rs64be,
		[Vuchar]=	Ru08be,
		[Vushort]=	Ru16be,
		[Vuint]=	Ru32be,
		[Vulong]=	Ru32be,
		[Vuvlong]=	Ru64be,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru32be,
	},
	.xtype = {
		[Xptr]=		Vulong,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vlong,
		[Xint64]=	Vvlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vulong,
		[Xuint64]=	Vuvlong,
	},
},
[Rc64le] = {
	.name = "c64le",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08le,
		[Vshort]=	Rs16le,
		[Vint]=		Rs32le,
		[Vlong]=	Rs64le,
		[Vvlong]=	Rs64le,
		[Vuchar]=	Ru08le,
		[Vushort]=	Ru16le,
		[Vuint]=	Ru32le,
		[Vulong]=	Ru64le,
		[Vuvlong]=	Ru64le,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru32le,
	},
	.xtype = {
		[Xptr]=		Vuint,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vint,
		[Xint64]=	Vlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vuint,
		[Xuint64]=	Vulong,
	},
},
[Rc64be] = {
	.name = "c64be",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08be,
		[Vshort]=	Rs16be,
		[Vint]=		Rs32be,
		[Vlong]=	Rs64be,
		[Vvlong]=	Rs64be,
		[Vuchar]=	Ru08be,
		[Vushort]=	Ru16be,
		[Vuint]=	Ru32be,
		[Vulong]=	Ru64be,
		[Vuvlong]=	Ru64be,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru32be,
	},
	.xtype = {
		[Xptr]=		Vuint,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vint,
		[Xint64]=	Vlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vuint,
		[Xuint64]=	Vulong,
	},
},
[Rclp64le] = {
	.name = "clp64le",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08le,
		[Vshort]=	Rs16le,
		[Vint]=		Rs32le,
		[Vlong]=	Rs64le,
		[Vvlong]=	Rs64le,
		[Vuchar]=	Ru08le,
		[Vushort]=	Ru16le,
		[Vuint]=	Ru32le,
		[Vulong]=	Ru64le,
		[Vuvlong]=	Ru64le,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru64le,
	},
	.xtype = {
		[Xptr]=		Vulong,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vint,
		[Xint64]=	Vlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vuint,
		[Xuint64]=	Vulong,
	},
},
[Rclp64be] = {
	.name = "clp64be",
	.base = {
		[Vbool]=	Ru08le,
		[Vchar]=	Rs08be,
		[Vshort]=	Rs16be,
		[Vint]=		Rs32be,
		[Vlong]=	Rs64be,
		[Vvlong]=	Rs64be,
		[Vuchar]=	Ru08be,
		[Vushort]=	Ru16be,
		[Vuint]=	Ru32be,
		[Vulong]=	Ru64be,
		[Vuvlong]=	Ru64be,
		[Vfloat]=	Rundef,
		[Vdouble]=	Rundef,
		[Vlongdouble]=	Rundef,
		[Vptr]=		Ru64be,
	},
	.xtype = {
		[Xptr]=		Vulong,
		[Xint8]=	Vchar,
		[Xint16]=	Vshort,
		[Xint32]=	Vint,
		[Xint64]=	Vlong,
		[Xuint8]=	Vuchar,
		[Xuint16]=	Vushort,
		[Xuint32]=	Vuint,
		[Xuint64]=	Vulong,
	},
}
};
