typedef
struct Dwarf
{
	char *abbrev;
	char *aranges;
	char *ehframe;
	char *frame;
	char *info;
	char *line;
	char *loc;
	char *pubnames;
	char *ranges;
	char *str;
	u64 vehframe;		/* vaddr of ehframe section */
	u32 nabbrev;
	u32 naranges;
	u32 nehframe;
	u32 nframe;
	u32 ninfo;
	u32 nline;
	u32 nloc;
	u32 npubnames;
	u32 nranges;
	u32 nstr;
	HT *cu;			/* offsets to compilation units */
	HT *sue;		/* named struct/union/enum tags to dies */
	HT *ant;		/* anonymous sue tags to dies */
	HT *typ;		/* type names to dies */
	HT *sym;		/* symbol names to dies */
	HT *xt;			/* extra types: type name to type */
	Type base[Vnallbase];	/* base types */
} Dwarf;

typedef
struct Nodwarf
{
	HT *xt;				/* extra types: type name to type */
	Type base[Vnallbase];		/* base types */
} Nodwarf;

struct Bin
{
	Bin *link;
	u32 refcnt;
	u64 id;
	u64 base;     			/* load address */
	u64 vtext;			/* text load address */
	u64 ntext;			/* size of text */
	char *path;			/* name of mapped binary file  */
	char *b;			/* mapping */
	u64 len;			/* size of mapping in bytes */
	unsigned char mode;		/* 32-bit or 64-bit */
	unsigned char enc;		/* little-endian or big-endian */
	unsigned char isexe;		/* is executable */
	char *symtab, *dynsym;		/* symbol tables */
	char *strtab, *dynstr;		/* corresponding string tables */
	char *relplt;			/* plt relocation table (rel format) */
	char *relaplt;			/* plt relocation table (rela format) */
	char *got;			/* got */
	u64 gotaddr;			/* got address */
	u32 nsym, ndyn;			/* table element counts */
	u32 nrel, nrela;		/* table element counts */
	u32 ndxdata, ndxrodata, ndxbss;	/* data sections indicies */
	u32 ndxtext;			/* text section index  */
	HT *sym;			/* symbols (including dwarf data) */
	Syms *ss;			/* sorted symbols */
	Dwarf dw;			/* dwarf goo */
	Nodwarf nodw;			/* types for dwarfless bins */
};

typedef
struct Seg
{
	u64 addr;
	u64 len;
	char *name;
} Seg;

typedef
struct Segs
{
	Seg *seg;
	u32 n, max;
} Segs;

/* dwarf unwind rules */

typedef
enum Rulekind
{
	Rnone = 0,
	// cfa rules
	Rcfaro,			/* cfa is register+offset */
	Rcfae,			/* unimplemented */
	// register rules
	Rsame,
	Roff,
	Rvoff,
	Rreg,
	Rexp,			/* unimplemented */
	Rvexp,			/* unimplemented */
} Rulekind;

enum
{
	Nreg = 67,		/* # dwarf regs in amd64 abi */
	CFA = Nreg,		/* index of CFA in Rule row */
	Nrow = Nreg+1,		/* # entries in Rule row */
};

typedef
struct Rule
{
	Rulekind kind;
	char *e;		/* dwarf expression */
	u64 r;			/* register */
	s64 n;			/* offset or length of dwarf expression */
} Rule;

extern int	defaultevents;
extern char*	dwarfattrname[];
extern char*	dwarfcfaname[];
extern char*	dwarfformname[];
extern char*	dwarfopname[];
extern char*	dwarftagname[];
extern char*	dwarflnsname[];

void		addbin(Ctl *ctl, Bin *b);
void		atbrkx86(Ctl *ctl);
char*		cidentifier(char *id);
void		dwarffini();
void		dwarfinit();
Types*		dwenumtype(Bin *b, char **err);
Type*		dwlooktype(Bin *b, Type *t, char **err);
Locs*		enumloc(Bin *b, u64 pc, char **err);
Syms*		enumsym(Bin *b);
Types*		enumtype(Bin *b, char **err);
Event*		eventsyscall(Ctl *ctl);
Event*		eventtrap(Ctl *ctl, Brkref *br);
Event*		eventsnap(Ctl *ctl, Brkref *br, Ctl *spwn);
Event*		eventspawn(Evkind kind, Ctl *ctl,  Ctl *spwn);
Event*		eventclone(Ctl *ctl, Ctl *spwn);
Event*		eventfork(Ctl *ctl, Ctl *spwn);
Event*		eventexit(Ctl *ctl, int st);
Event*		eventexec(Ctl *ctl);
Event*		eventload(Ctl *ctl, Bin **bs);
Event*		eventsignal(Ctl *ctl, siginfo_t *info);
void		freedwarf(Bin *b);
void		freebin(Bin *b);
void		freelocs(Locs *ls);
void		freesym(Sym *s);
void		freetypes(Types *t);
void		initdwarf(Bin *b);
char*		interp(char *exe, u8 *m, char **err);
int		ldevent(Ctl *ctl, Bin ***bs, char **err);
u64		loadaddr(char *exe, char **err);
Bin*		lookbinpath(Ctl *ctl, char *path);
Bin*		lookbinpc(Ctl *ctl, u64 pc);
Sym*		looksym(Bin *b, char *id);
Type*		looktype(Bin *b, Type *t, char **err);
Bin*		mkbin(char *file, u64 base, char **err);
Ctl*		mkctl(u32 pid);
Sym*		mksym();
Type*		mkvoid();
Bin*		names(char *path, char **err);
void		packsiginfo(Buf *b, siginfo_t *info);  /* FIXME: belongs elsewhere? */
int		parsemaps(Stat *s, char **err);
int		resetexec(Ctl *ctl, char **err);
void		symfreeas(As *as);
void		symfini();
void		syminit();
int		syminitattach(Ctl *ctl, char **err);
int		syminitexec(Ctl *ctl, char **err);
int		symresetexec(Ctl *ctl, char **err);
void		symstat(Ctl *ctl, Stat *s);
void		tvdiff(struct timeval *a, struct timeval *b, struct timeval *c);
int		unwind1(Bin *b, u64 pc, Rule *rule, char **err);
void		xgetregs(Ctl *ctl, Reg *regs);
void		xsetregs(Ctl *ctl, Reg *regs);
