#include "../inc/sctl.h"

void
killall()
{
	// FIXME: valgrind doesn't get a chance if we exit this way.
	kill(0, SIGKILL);
	abort();
}

char*
lasterror()
{
	return strerror(errno);
}

void
cloexec(int fd)
{
	int f;
	f = fcntl(fd, F_GETFD);
	if(f == -1)
		xfatal("fcntl: %s", strerror(errno));
	f |= FD_CLOEXEC;
	if(fcntl(fd, F_SETFD, f))
		xfatal("fcntl: %s", strerror(errno));
}

char*
readfile(char *name, u64 *plen)
{
	char *rv;
	u64 len, max, munch;
	int fd;
	int m;

	len = 0;
	max = 1024;
	rv = xmalloc(max+1);

	fd = open(name, O_RDONLY);
	if(0 > fd)
		goto fail;

	while(1){
		if(len >= max){
			rv = xrealloc(rv, max+1, max*2);
			max *= 2;
		}
		munch = max-len;
		m = read(fd, rv+len, munch);
		if(0 > m && (errno == EINTR || errno == EAGAIN))
			continue;
		if(0 > m)
			goto fail;
		len += m;
		if(m < munch)
			break;
	}
	close(fd);
	rv[len] = 0;
	if(plen)
		*plen = len;
	return rv;
fail:
	close(fd);
	free(rv);
	return 0;
}

s64
xread(Fd fd, void *buf, size_t len)
{
	char *p = buf;
	size_t nrecv = 0;
	ssize_t rv;

	while(nrecv < len) {
		rv = read(fd, p, len-nrecv);
		if(0 > rv && (errno == EINTR || errno == EAGAIN))
			continue;
		if(0 > rv)
			return -1;
		if(0 == rv)
			return nrecv;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

s64
xrecv(Chan sd, void *buf, size_t len)
{
	char *p = buf;
	size_t nrecv = 0;
	ssize_t rv;

	while(nrecv < len) {
		rv = recv(sd, p, len-nrecv, 0);
		if(0 > rv && (errno == EINTR || errno == EAGAIN))
			continue;
		if(0 > rv)
			return -1;
		if(0 == rv)
			return nrecv;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

void
xsend(Chan sd, void *buf, size_t len)
{
	char *p = buf;
	size_t nsent = 0;
	ssize_t rv;

	while (nsent < len) {
		rv = send(sd, p, len-nsent, 0);
		if(0 > rv && (errno == EINTR || errno == EAGAIN))
			continue;
		if (0 > rv)
			xfatalno("write");
		if (0 == rv)
			xfatalno("write");
		nsent += rv;
		p += rv;
	}
}
