#include "../inc/sctl.h"
#include "../inc/err.h"

enum
{
	Maxidle = 60,		/* seconds */
};

static void
usage(void)
{
	printf("usage: sctl [options]\n");
	printf("  options:\n");
	printf("     -p <port>  listen for client on <port>\n");
	printf("     -d         increase debug level\n");
	printf("\n");
	exit(0);
}

int
main(int argc, char *argv[])
{
	unsigned short port;
	Chan fd, c;
	struct sockaddr_in sa, ca;
	socklen_t len;
	int op, rv;
	fd_set fds;
	struct timeval tv;
	Chan parent_sock;
	WSADATA wsaData;

	setargv0(argv[0]);

	if(argc < 2)
		usage();

	port = 0;
	parent_sock = INVALID_SOCKET;
	ARGBEGIN{
	case 'p':
		port = (unsigned short)atoi(EARGF(usage()));
		break;
	case 'd':
		debuglevel++;
		break;
	case 'r':
		parent_sock = (SOCKET)atoi(EARGF(usage()));
		break;
	default:
		usage();

	}ARGEND;

	if(WSAStartup(0x0202,&wsaData))
		xfatalno("WSAStartup");

	if(parent_sock != INVALID_SOCKET){
		ioread(parent_sock);
		exit(0);
	}

	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons(port);

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(0 > fd)
		xfatalno("socket");
	reuseaddr(fd);
	if(0 > bind(fd, (struct sockaddr*)&sa, sizeof(sa)))
		xfatalno("bind");
	if(0 > listen(fd, 1))
		xfatalno("listen");

	while(1){
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		memset(&tv, 0, sizeof(tv));
		tv.tv_sec = Maxidle;
		rv = select((int)fd+1, &fds, 0, 0, &tv);
		if(rv < 0)
			xfatalno("select");
		if(rv == 0)
			break;
		len = sizeof(ca);
		c = accept(fd, (struct sockaddr*)&ca, &len);
		if(INVALID_SOCKET == c)
			xfatalno("accept");
		nodelay(c);
		prctl(c);
	}
	return 0;
}
