#include "../inc/sctl.h"
#include "../inc/err.h"
#include "win.h"

void
freeps(Ps *p)
{
	u64 i;
	for(i = 0; i < p->npr; i++)
		free(p->pr[i].cmd);
	free(p->pr);
	free(p);
}

void
packps(Buf *b, Ps *p)
{
	u64 i, j;
	Proc *pr;
	Thread *th;

	bufpushu64(b, p->npr);
	for(i = 0; i < p->npr; i++){
		pr = &p->pr[i];
		bufpushu64(b, pr->pid);
		bufpushs(b, pr->cmd);
		bufpushu64(b, pr->nth);
		for(j = 0; j < pr->nth; j++){
			th = &pr->th[j];
			bufpushu64(b, th->tid);
		}
	}
}

static char*
procname(int pid, char **err)
{
	char path[MAX_PATH];
	HANDLE p;
        HMODULE *mods;
	MODULEINFO mi;
        DWORD max, need, n, i;

	p = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ,
			FALSE, pid);
	if(p == 0)
		goto fail;

	max = 4096*sizeof(HMODULE);
again:
	mods = xmalloc(max);
	if(!EnumProcessModules(p, mods, max, &need)){
		free(mods);
		if(err)
			*err = lasterror();
		return 0;
	}
	if(need > max){
		max = need;
		free(mods);
		goto again;
	}
	n = need/sizeof(HMODULE);
	for(i = 0; i < n; i++){
		path[0] = 0;
		GetModuleBaseName(p, mods[i], path, sizeof(path));
		GetModuleInformation(p, mods[i], &mi, sizeof(mi));
	}
	path[0] = 0;
	GetModuleBaseName(p, mods[0], path, sizeof(path));
	free(mods);
	CloseHandle(p);
	if(path[0])
		return xstrdup(path);
fail:
	return xstrdup("<unknown>");
}

Ps*
ps(char **err)
{
	DWORD max, need, *pids;
	u32 n, i;
	u64 nr;
	char *cmd;
	Ps *p;
	HANDLE pr;
	Region *r;

	max = 4096*sizeof(DWORD);
again:
	pids = xmalloc(max);
	if(!EnumProcesses(pids, max, &need)){
		free(pids);
		if(err)
			*err = lasterror();
		return 0;
	}
	if(need > max){
		max = need;
		free(pids);
		goto again;
	}

	n = need/sizeof(DWORD);
	p = xmalloc(sizeof(Ps));
	p->pr = xmalloc(n*sizeof(Proc));
	p->npr = n;
	for(i = 0; i < n; i++){
		pr = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ,
				 FALSE, pids[i]);
		if(pr == 0){
			cmd = xstrdup("unknown");
			goto fail;
		}
		r = dlls(pr, &nr, err);
		CloseHandle(pr);
		if(r == 0){
			cmd = xstrdup("unknown");
			goto fail;
		}
		if(nr)
			cmd = xstrdup(r[0].path);
		else
			cmd = xstrdup("unknown");
		freeregions(r, nr);
fail:
		p->pr[i].cmd = cmd;
		p->pr[i].pid = pids[i];
		p->pr[i].nth = 0;
	}

	free(pids);
	return p;
}
