#include "../inc/sctl.h"
#include "queue.h"

typedef struct __item {
	void *v;
	struct __item *next;
} Q_item;

struct Q {
	Q_item *head;
	Q_item *tail;
};

Q*
mkq()
{
	Q *newq = xmalloc(sizeof(Q));
	newq->head=NULL;
	newq->tail=NULL;

	return newq;
}

void
qadd(Q *q, void *v)
{
	Q_item *qi = xmalloc(sizeof(Q_item));
	qi->v = v;
	qi->next = NULL;

	if(q->tail == NULL)
		q->tail = qi;
	else {
		q->tail->next = qi;
		q->tail = qi;
	}

	if(q->head == NULL)
		q->head = q->tail;
}

static void
wrapadd(void *u, void *v)
{
	Q *dst;
	dst = u;
	qadd(dst, v);
}
void
qaddall(Q *dst, Q *src)
{
	qforeach(src, wrapadd, dst);
}

void
qpush(Q *q, void *v)
{
	Q_item *qi = xmalloc(sizeof(Q_item));
	qi->v = v;

	qi->next = q->head;
	q->head = qi;

	if(q->tail == NULL)
		q->tail = q->head;
}

void *
qdeq(Q *q)
{
	Q_item *qi;
	void *v;

	if(q->head) {
		qi = q->head;
		v = qi->v;
		q->head = qi->next;

		if(!q->head)
			q->tail = NULL;

		free(qi);
		return v;
	} else
		return NULL;
}

void
qforeach(Q *q, void (*f)(void *u, void *v), void *u)
{
	Q_item *qi;
	qi = q->head;

	while(qi) {
		f(u, qi->v);
		qi = qi->next;
	}
}

unsigned
qsize(Q *q)
{
	unsigned n;
	Q_item *qi;

	n = 0;
	qi = q->head;

	while(qi) {
		n++;
		qi = qi->next;
	}
	return n;
}

void
qarrayof(Q *q, void** arr, unsigned len)
{
	unsigned i;
	Q_item *qi;

	i = 0;
	qi = q->head;

	while(qi && i < len) {
		arr[i++] = qi->v;
		qi = qi->next;
	}
}

void
qfree(Q *q) {

	while(q->head) {
		Q_item *qi = q->head;
		q->head = q->head->next;
		free(qi);
	}
	free(q);
}


