#include "../inc/sctl.h"
#include "../inc/err.h"
#include "win.h"

void
ioread(Chan fd)
{
	fd_set fds;
	while(1){
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		if(SOCKET_ERROR == select((int)fd+1, &fds, 0, 0, 0))
			xfatalno("ioread");
		DebugBreak();
	}
}

/*
 * multiplex DEBUG_EVENTs and rpcs
 * by converting rpc arrivals into DEBUG_EVENTs
 */
void
prctl(Chan fd)
{
	int iopid, dc;
	char argbuf[MAX_PATH+1];
	char *err;
	Msg *m;
	STARTUPINFO si = {sizeof(si)};
	PROCESS_INFORMATION pi;
	HANDLE dup_fd;
	DEBUG_EVENT dbg, *nextdbg;
	Event *e;

	sctlinit();

	/* FIXME: use SetHandleInformation to remove HANDLE_FLAG_INHERIT
	 *   from handles we don't want to pass on to ioread */

	/* Duplicate the socket to create an inheritable copy */
	if(!DuplicateHandle(GetCurrentProcess(),
			    (HANDLE)fd,
			    GetCurrentProcess(),
			    (HANDLE*)&dup_fd,
			    0,
			    TRUE, // Inheritable
			    DUPLICATE_SAME_ACCESS))
		xfatalno("DuplicateHandle");

	/* Create ioread helper (inherits socket handle) */
	sprintf_s(argbuf, sizeof(argbuf), "%s -r %d", argv0, (int)dup_fd);

	if(!CreateProcess(NULL,
			  argbuf, NULL, NULL,
			  TRUE, // Inherit handles
			  DEBUG_ONLY_THIS_PROCESS,
			  NULL, NULL, &si, &pi))
		xfatalno("CreateProcess");

	/* wait for ioread proc to begin */
	do{
		if(!WaitForDebugEvent(&dbg, INFINITE))
			xfatalno("WaitForDebugEvent");
		dc = dbg.dwDebugEventCode;
		switch(dc){
		case LOAD_DLL_DEBUG_EVENT:
			if(dbg.u.LoadDll.hFile)
				CloseHandle(dbg.u.LoadDll.hFile);
			break;
		case CREATE_PROCESS_DEBUG_EVENT:
		case EXCEPTION_DEBUG_EVENT:
			break;
		default:
			xfatal("unexpected %s", dbg2str(dc));
		}
		if(!ContinueDebugEvent(dbg.dwProcessId,
				       dbg.dwThreadId,
				       DBG_CONTINUE))
			xfatalno("ContinueDebugEvent");
	}while(dc != EXCEPTION_DEBUG_EVENT);

	iopid = pi.dwProcessId;
	while(1){
		while(e = evqpop()){
			aevent(fd, e);
			freeevent(e);
		}
		if(!WaitForDebugEvent(&dbg, INFINITE))
			xfatalno("WaitForDebugEvent");
		if(dbg.dwProcessId == iopid){
			dc = dbg.dwDebugEventCode;
			switch(dc){
			case EXCEPTION_DEBUG_EVENT:
				m = readmsg(fd);
				if(m == 0){
					closechan(fd);
					sctlfini();
					killall();
				}
				domsg(fd, m);
				freemsg(m);
				break;
			case LOAD_DLL_DEBUG_EVENT:
				if(dbg.u.LoadDll.hFile)
					CloseHandle(dbg.u.LoadDll.hFile);
				break;
			default:
				printf("unexpected %s\n", dbg2str(dc));
				bug();
				break;
			}
			if(!ContinueDebugEvent(dbg.dwProcessId,
					       dbg.dwThreadId,
					       DBG_CONTINUE))
				xfatalno("ContinueDebugEvent");
		}else{
			err = 0;
			e = doevent(&dbg, &err);
			if(!e && err)
				xfatal("doevent: %s\n", err);
			if(e){
				aevent(fd, e);
				freeevent(e);
			}
		}
	}
}
