#include "../inc/sctl.h"
#include "../inc/err.h"
#include "win.h"

enum
{
	LIM = 4096,
};

void
freestat(Stat *s)
{
	freeregions(s->rs, s->nr);
	free(s->ts);
	free(s);
}

void
packstat(Buf *b, Stat *s)
{
	u64 i;

	bufpushu32(b, s->state);

	bufpushu64(b, s->nr);
	for(i = 0; i < s->nr; i++){
		bufpushs(b, s->rs[i].path);
		bufpushu64(b, s->rs[i].base);
	}

	bufpushu64(b, s->nt);
	for(i = 0; i < s->nt; i++){
		bufpushu64(b, s->ts[i].id);
		bufpushu32(b, s->ts[i].state);
	}
}

void
freeregions(Region *r, u32 nr)
{
	u32 i;
	for(i = 0; i < nr; i++)
		free(r[i].path);
	free(r);
}

Region*
dlls(HANDLE p, u64 *nr, char **err)
{
	char path[MAX_PATH];
        HMODULE *mods;
	MODULEINFO mi;
        DWORD max, need, n, i;
	Region *r;

	max = 4096*sizeof(HMODULE);
again:
	mods = xmalloc(max);
	if(!EnumProcessModules(p, mods, max, &need)){
		free(mods);
		if(err)
			*err = lasterror();
		return 0;
	}
	if(need > max){
		max = need;
		free(mods);
		goto again;
	}
	n = need/sizeof(HMODULE);
	r = xmalloc(n*sizeof(Region));
	for(i = 0; i < n; i++){
		path[0] = 0;
//		GetModuleBaseName(p, mods[i], path, sizeof(path));
		GetModuleFileNameEx(p, mods[i], path, sizeof(path));
		GetModuleInformation(p, mods[i], &mi, sizeof(mi));
		r[i].path = xstrdup(path);
		r[i].base = (u64)mi.lpBaseOfDll;
	}
	free(mods);
	*nr = (u64)n;
	return r;
}

Stat*
ctlstat(Ctl *ctl, char **err)
{
	Stat *s;
	Ctl *p;
	u64 n;

	s = xmalloc(sizeof(Stat));
	s->state = ctl->state == Srun ? 0 : 1;

	s->rs = dlls(ctl->x.process, &s->nr, err);
	if(s->rs == 0){
		free(s);
		return 0;
	}

	n = 0;
	p = ctl->as->ctl;
	while(p){
		n++;
		p = p->link;
	}
	s->ts = xmalloc(n*sizeof(*s->ts));
	p = ctl->as->ctl;
	while(p){
		s->ts[s->nt].id = p->id;
		s->ts[s->nt].state = p->state == Srun ? 0 : 1;
		s->nt++;
		p = p->link;
	}

	return s;
}
