#include "../inc/sctl.h"

char*
lasterror()
{
	/* not thread safe? */
	int err;
	static void *lpMsgBuf = 0;
	if(lpMsgBuf != 0)
		LocalFree(lpMsgBuf);
	err = GetLastError();
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM,
		0,
		err,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0,
		0
		);
	return (char*) lpMsgBuf;
}

void
killall()
{
	/* stderr is buffered by default on windows */
	fflush(stderr);
	/* kills all attached processes */
	TerminateProcess(GetCurrentProcess(),1);
}

s64
xread(Fd fd, void *buf, size_t len)
{
	char *p = buf;
	size_t nrecv = 0;
	ssize_t rv;

	while(nrecv < len) {
		rv = read(fd, p, len-nrecv);
		if(0 > rv)
			return -1;
		if(0 == rv)
			return nrecv;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

s64
xrecv(Chan sd, void *buf, size_t len)
{
	char *p = buf;
	size_t nrecv = 0;
	ssize_t rv;

	while(nrecv < len) {
		rv = recv(sd, p, len-nrecv, 0);
		if(0 > rv && (WSAGetLastError()== EINTR
			      || WSAGetLastError() == EAGAIN))
			continue;
		if(0 > rv)
			return -1;
		if(0 == rv)
			return nrecv;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

void
xsend(Chan sd, void *buf, size_t len)
{
	char *p = buf;
	size_t nsent = 0;
	ssize_t rv;

	while (nsent < len) {
		rv = send(sd, p, len-nsent, 0);
		if (0 > rv && (WSAGetLastError()== EINTR
			       || WSAGetLastError() == EAGAIN))
			continue;
		if (0 > rv)
			xfatalno("write");
		if (0 == rv)
			xfatalno("write");
		nsent += rv;
		p += rv;
	}
}

void
dumpstack()
{
}
