#include "../inc/sctl.h"
#include "../inc/err.h"

typedef
struct Packed {
	u64 r15;
	u64 r14;
	u64 r13;
	u64 r12;
	u64 rbp;
	u64 rbx;
	u64 r11;
	u64 r10;
	u64 r9;
	u64 r8;
	u64 rax;
	u64 rcx;
	u64 rdx;
	u64 rsi;
	u64 rdi;
	u64 orig_rax;
	u64 rip;
	u64 cs;
	u64 eflags;
	u64 rsp;
	u64 ss;
	u64 fs_base;
	u64 gs_base;
	u64 ds;
	u64 es;
	u64 fs;
	u64 gs;

	/* current signal number */
	u64 sig;
} Packed;

u64
ctlpc(Ctl *ctl)
{
	return ctl->ctx.regs.Eip;
}

void
ctlsetpc(Ctl *ctl, u64 pc)
{
	ctl->ctx.regs.Eip = pc;
}

void
freectx(Ctx *c)
{
	free(c);
}

Ctx*
unpackctx(void *bytes, u64 n)
{
	Packed *p;
	Ctx *ctx;

	if(n != sizeof(Packed))
		return 0;
	p = bytes;
	ctx = xmalloc(sizeof(Ctx));
	ctx->regs.EFlags = p->eflags;
	ctx->regs.Ebx = p->rbx;
	ctx->regs.Ecx = p->rcx;
	ctx->regs.Edx = p->rdx;
	ctx->regs.Esi = p->rsi;
	ctx->regs.Edi = p->rdi;
	ctx->regs.Ebp = p->rbp;
	ctx->regs.Eax = p->rax;
	ctx->regs.Eip = p->rip;
	ctx->regs.Esp = p->rsp;
	ctx->curexn = p->sig;
	return ctx;
}

void
packctx(Buf *ob, Ctx *ctx)
{
	Packed p;

	memset(&p, 0, sizeof(Packed));

	p.eflags = ctx->regs.EFlags;
	p.rbx = ctx->regs.Ebx;
	p.rcx = ctx->regs.Ecx;
	p.rdx = ctx->regs.Edx;
	p.rsi = ctx->regs.Esi;
	p.rdi = ctx->regs.Edi;
	p.rbp = ctx->regs.Ebp;
	p.rax = ctx->regs.Eax;
	p.rip = ctx->regs.Eip;
	p.rsp = ctx->regs.Esp;
	p.sig = ctx->curexn;

	bufpush(ob, &p, sizeof(Packed));
}
